# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import re
import tempfile
from typing import Any, Dict, List, cast

import rpm  # type: ignore

from rebasehelper.constants import SYSTEM_ENCODING
from rebasehelper.exceptions import RebaseHelperError
from rebasehelper.logger import CustomLogger
from rebasehelper.helpers.macro_helper import MacroHelper
from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.console_helper import ConsoleHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class RpmHeader:
    def __init__(self, hdr: rpm.hdr) -> None:
        self.hdr = hdr

    def __getattr__(self, item: str) -> Any:
        def decode(s):
            if isinstance(s, bytes):
                return s.decode(SYSTEM_ENCODING)
            return s
        result = getattr(self.hdr, item)
        if isinstance(result, list):
            return [decode(x) for x in result]
        return decode(result)


class RpmHelper:

    """Class for working with RPM database and packages."""

    ARCHES: List[str] = []

    @staticmethod
    def is_package_installed(pkg_name=None):
        """Checks whether a package is installed.

        Args:
            pkg_name (str): Name of the package.

        Returns:
            bool: Whether the package is installed.

        """
        ts = rpm.TransactionSet()
        mi = ts.dbMatch('provides', pkg_name)
        return len(mi) > 0

    @staticmethod
    def all_packages_installed(pkg_names=None):
        """Checks if all specified packages are installed.

        Args:
            pkg_names (list): List of package names to check.

        Returns:
            bool: True if all packages are installed, False otherwise.

        """
        for pkg in pkg_names:
            if not RpmHelper.is_package_installed(pkg):
                return False
        return True

    @staticmethod
    def install_build_dependencies(spec_path=None, assume_yes=False):
        """Installs build dependencies of a package using dnf.

        Args:
            spec_path (str): Absolute path to the SPEC file.
            assume_yes (bool): Whether to automatically answer 'yes' to all questions.

        Returns:
            int: Exit code of the subprocess run.

        """
        cmd = ['dnf', 'builddep', spec_path]
        if os.geteuid() != 0:
            logger.warning("Authentication required to install build dependencies using '%s'", ' '.join(cmd))
            cmd = ['pkexec'] + cmd
        if assume_yes:
            cmd.append('-y')
        return ProcessHelper.run_subprocess(cmd)

    @staticmethod
    def get_header_from_rpm(rpm_name):
        """Gets an RPM header from the given RPM package.

        Args:
            rpm_name (str): Path to the package.

        Returns:
            RpmHeader: Header object obtained from the package.

        """
        ts = rpm.TransactionSet()
        # disable signature checking
        ts.setVSFlags(rpm._RPMVSF_NOSIGNATURES)  # pylint: disable=protected-access
        with open(rpm_name, "r") as f:
            hdr = ts.hdrFromFdno(f)
        return RpmHeader(hdr)

    @staticmethod
    def get_arches():
        """Gets list of all known architectures"""
        arches = ['aarch64', 'noarch', 'ppc', 'riscv64', 's390', 's390x', 'src', 'x86_64']
        macros = MacroHelper.dump()
        macros = [m for m in macros if m['name'] in ('ix86', 'arm', 'mips', 'sparc', 'alpha', 'power64')]
        for m in macros:
            arches.extend(MacroHelper.expand(m['value'], '').split())
        return arches

    @classmethod
    def split_nevra(cls, s):
        """Splits string into name, epoch, version, release and arch components"""
        regexps = [
            ('NEVRA', re.compile(r'^([^:]+)-(([0-9]+):)?([^-:]+)-(.+)\.([^.]+)$')),
            ('NEVR', re.compile(r'^([^:]+)-(([0-9]+):)?([^-:]+)-(.+)()$')),
            ('NA', re.compile(r'^([^:]+)()()()()\.([^.]+)$')),
            ('N', re.compile(r'^([^:]+)()()()()()$')),
        ]
        if not cls.ARCHES:
            cls.ARCHES = cls.get_arches()
        for pattern, regexp in regexps:
            match = regexp.match(s)
            if not match:
                continue
            name = match.group(1) or None
            epoch = match.group(3) or None
            if epoch:
                epoch = int(epoch)
            version = match.group(4) or None
            release = match.group(5) or None
            arch = match.group(6) or None
            if pattern == 'NEVRA' and arch not in cls.ARCHES:
                # unknown arch, let's assume it's actually dist
                continue
            return dict(name=name, epoch=epoch, version=version, release=release, arch=arch)
        raise RebaseHelperError('Unable to split string into NEVRA.')

    @classmethod
    def parse_spec(cls, path, flags=None):
        with open(path, 'rb') as orig:
            with tempfile.NamedTemporaryFile() as tmp:
                # remove BuildArch to workaround rpm bug
                tmp.write(b''.join(l for l in orig.readlines() if not l.startswith(b'BuildArch')))
                tmp.flush()
                capturer = None
                try:
                    with ConsoleHelper.Capturer(stderr=True) as capturer:
                        result = rpm.spec(tmp.name, flags) if flags is not None else rpm.spec(tmp.name)
                except ValueError as e:
                    output = capturer.stderr.strip().split('\n') if capturer else []
                    if len(output) == 1:
                        output = output[0]
                    raise RebaseHelperError('Failed to parse SPEC file{0}'.format(
                        ': ' + str(output) if output else '')) from e
                return result

    @classmethod
    def get_rpm_spec(cls, path: str, sourcedir: str, predefined_macros: Dict[str, str]) -> rpm.spec:
        # reset all macros and settings
        rpm.reloadConfig()
        # ensure that %{_sourcedir} macro is set to proper location
        MacroHelper.purge_macro('_sourcedir')
        rpm.addMacro('_sourcedir', sourcedir)
        # add predefined macros
        for macro, value in predefined_macros.items():
            rpm.addMacro(macro, value)
        try:
            spec = cls.parse_spec(path, flags=rpm.RPMSPEC_ANYARCH)
        except RebaseHelperError:
            # try again with RPMSPEC_FORCE flag (the default)
            spec = cls.parse_spec(path)
        return spec



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import io
import os
import re
from typing import Dict, List, Optional

from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.plugins.checkers import BaseChecker, CheckerCategory


class LicenseCheck(BaseChecker):
    """license compare tool

    Attributes:
        license_changes(bool): Boolean value to inform whether any license change occured.
        license_files_changes(dict): Dictionary of {license: change_type + file_name} pairs.
    """

    DEFAULT: bool = True
    CATEGORY: Optional[CheckerCategory] = CheckerCategory.SOURCE

    CMD: str = 'licensecheck'
    license_changes: bool = False
    license_files_changes: Dict[str, str] = {}

    @classmethod
    def is_available(cls):
        try:
            return ProcessHelper.run_subprocess([cls.CMD, '--help'], output_file=ProcessHelper.DEV_NULL) == 0
        except (IOError, OSError):
            return False

    @classmethod
    def get_license_changes(cls, old_dir, new_dir):
        """
        Finds differences in licenses between old and new source files.

        Args:
            old_dir(str): Path to the old sources directory.
            new_dir(str): Path to the new sources directory.

        Returns:
            tuple: Changes dictionary, new_licenses set, disappeared_licenses set.
        """
        diffs = []
        possible_license_names = r'COPYING|LICENSE|LICENCE|LICENSING|BSD|(L)?GPL(v[23][+]?)?'
        for source_dir in [old_dir, new_dir]:
            out = io.StringIO()
            ProcessHelper.run_subprocess(["/usr/bin/licensecheck", source_dir, "--machine", "--recursive"],
                                         output_file=out, ignore_stderr=True)
            diff = {}
            for l in out:
                # licensecheck output format: 'Filepath\tlicense'
                file_path, dlicense = l.split('\t')
                if 'GENERATED FILE' in dlicense:
                    continue
                file_path = os.path.relpath(file_path, source_dir)
                diff[file_path] = dlicense.strip()
            diffs.append(diff)

        old_lics, new_lics = set(), set()
        changes: Dict[str, Dict[str, List[str]]] = {'appeared': {}, 'transitioned': {}, 'disappeared': {}}
        # Get changed licenses in existing files
        for new_file, new_license in diffs[1].items():
            new_lics.add(new_license)
            for old_file, old_license in diffs[0].items():
                old_lics.add(old_license)

                if (new_file == old_file and
                   (new_license != old_license)):
                    new_key = '{} => {}'.format(old_license, new_license)
                    if new_license == 'UNKNOWN':
                        # Conversion `known license` => `None/Unknown`
                        if re.search(possible_license_names, new_file, re.IGNORECASE):
                            cls.license_files_changes.update({old_license: 'disappeared in {}'.format(new_file)})
                        if old_license not in changes['disappeared']:
                            changes['disappeared'][old_license] = []
                        changes['disappeared'][old_license].append(new_file)
                    elif old_license == 'UNKNOWN':
                        # Conversion `None/Unknown` => `known license`
                        if re.search(possible_license_names, new_file, re.IGNORECASE):
                            cls.license_files_changes.update({new_license: 'appeared in {}'.format(new_file)})
                        if new_license not in changes['appeared']:
                            changes['appeared'][new_license] = []
                        changes['appeared'][new_license].append(new_file)
                    else:
                        # Conversion `known license` => `known license`
                        if re.search(possible_license_names, new_file, re.IGNORECASE):
                            cls.license_files_changes.update({new_key: 'in {}'.format(new_file)})
                        if new_key not in changes['transitioned']:
                            changes['transitioned'][new_key] = []
                        if new_file not in changes['transitioned'][new_key]:
                            changes['transitioned'][new_key].append(new_file)

        # Get newly appeared files
        for new_file, new_license in diffs[1].items():
            if new_file not in diffs[0]:
                if new_license == 'UNKNOWN':
                    continue
                if re.search(possible_license_names, new_file, re.IGNORECASE):
                    cls.license_files_changes.update({new_license: 'appeared in {}'.format(new_file)})
                if new_license not in changes['appeared']:
                    changes['appeared'][new_license] = []
                changes['appeared'][new_license].append(new_file)

        # Get removed files
        for old_file, old_license in diffs[0].items():
            if old_file not in diffs[1]:
                if old_license == 'UNKNOWN':
                    continue
                if re.search(possible_license_names, old_file, re.IGNORECASE):
                    cls.license_files_changes.update({old_license: 'disappeared in {}'.format(old_file)})
                if old_license not in changes['disappeared']:
                    changes['disappeared'][old_license] = []
                changes['disappeared'][old_license].append(old_file)

        new_licenses = new_lics - old_lics
        disappeared_licenses = old_lics - new_lics
        if new_licenses or disappeared_licenses:
            cls.license_changes = True

        return changes, list(new_licenses), list(disappeared_licenses)

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        cls.license_changes = False
        cls.license_files_changes = dict()
        cls.results_dir = os.path.join(results_dir, 'licensecheck')
        cls.prepare_results_dir()
        changes, new_licenses, disappeared_licenses = cls.get_license_changes(kwargs['old_dir'], kwargs['new_dir'])
        cls.output_to_report_file(changes, os.path.join(cls.results_dir, 'report.txt'))

        return {'path': cls.get_checker_output_dir_short(), 'changes': changes,
                'license_changes': cls.license_changes, 'new_licenses': new_licenses,
                'disappeared_licenses': disappeared_licenses, 'license_files_changes': cls.license_files_changes}

    @classmethod
    def output_to_report_file(cls, changes, report_file_path):
        """
        Prints the licensecheck output to a report file.

        Args:
            changes(dict): Changes dictionary produced by licensecheck_tool.
            report_file_path(str): Path for the report file.
        """
        output_string = [cls.get_underlined_title("licensecheck").lstrip()]
        if cls.license_changes:
            output_string.append('License changes occured!')
            for change_name, change_info in sorted(changes.items()):
                if not change_info:
                    continue
                output_string.append(cls.get_underlined_title('The following license(s) {}'.format(change_name)))
                for license_name, files in sorted(change_info.items()):
                    output_string.append('* {}'.format(license_name))
                    for f in sorted(files):
                        output_string.append(' - {}'.format(f))
        else:
            output_string.append('No license changes detected.')

        with open(report_file_path, 'w') as f:
            f.write('\n'.join(output_string))

    @classmethod
    def format(cls, data):
        output_string = [cls.get_underlined_title("licensecheck")]

        if data['license_changes']:
            if data['license_files_changes']:
                output_string.append('Major license changes:')
                for license_change, change_type in cls.license_files_changes.items():
                    output_string.append(' - {} {}'.format(license_change, change_type))
                    # Remove already mentioned transition to avoid duplicity in the report
                    if license_change in data['new_licenses']:
                        data['new_licenses'].remove(license_change)
                    if license_change in data['disappeared_licenses']:
                        data['disappeared_licenses'].remove(license_change)

            output_string.append('Other license changes:')
            for l in sorted(data['new_licenses']):
                output_string.append(' - {} appeared'.format(l))
            for l in sorted(data['disappeared_licenses']):
                output_string.append(' - {} disappeared'.format(l))

            output_string.append('Details in {}:'.format(data['path']))
            output_string.append(' - {}'.format('report.txt'))
        else:
            output_string.append('No license changes detected.')

        return output_string

    @classmethod
    def get_important_changes(cls, checker_output):
        if LicenseCheck.license_changes:
            return ['License changes occured. Check licensecheck output.']
        return []



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os
import shutil

import git  # type: ignore
import pytest  # type: ignore

from typing import List

from rebasehelper.patcher import Patcher
from rebasehelper.specfile import PatchObject


class TestPatcher:

    USER: str = 'John Doe'
    EMAIL: str = 'john.doe@example.com'

    LIPSUM_OLD: str = 'patcher/lipsum_old.txt'
    LIPSUM_NEW: str = 'patcher/lipsum_new.txt'
    PATCH1: str = 'patcher/1.patch'
    PATCH2: str = 'patcher/2.patch'
    PATCH3: str = 'patcher/3.patch'
    PATCH4: str = 'patcher/4.patch'

    TEST_FILES: List[str] = [
        LIPSUM_OLD,
        LIPSUM_NEW,
        PATCH1,
        PATCH2,
        PATCH3,
        PATCH4,
    ]

    @pytest.fixture
    def rebased_sources(self, workdir):
        path = os.path.join(workdir, 'rebased_sources')
        os.mkdir(path)
        return path

    @pytest.fixture
    def old_sources(self, workdir):
        path = os.path.join(workdir, 'old_sources')
        os.mkdir(path)
        return path

    @pytest.fixture
    def new_sources(self, workdir):
        path = os.path.join(workdir, 'new_sources')
        os.mkdir(path)
        return path

    @pytest.fixture
    def old_repo(self, old_sources):
        repo = git.Repo.init(old_sources)
        repo.git.config('user.name', self.USER, local=True)
        repo.git.config('user.email', self.EMAIL, local=True)
        shutil.copy(os.path.basename(self.LIPSUM_OLD), os.path.join(old_sources, 'lipsum.txt'))
        repo.git.add(all=True)
        repo.index.commit('Initial commit', skip_hooks=True)
        for n in range(1, 5):
            patch_name = os.path.basename(getattr(self, 'PATCH{0}'.format(n)))
            repo.git.apply(os.path.join(os.getcwd(), patch_name))
            repo.git.add(all=True)
            repo.index.commit(Patcher.insert_patch_name('P{0}'.format(n), patch_name), skip_hooks=True)
        return repo

    @pytest.fixture
    def new_repo(self, new_sources):
        repo = git.Repo.init(new_sources)
        repo.git.config('user.name', self.USER, local=True)
        repo.git.config('user.email', self.EMAIL, local=True)
        shutil.copy(os.path.basename(self.LIPSUM_NEW), os.path.join(new_sources, 'lipsum.txt'))
        repo.git.add(all=True)
        repo.index.commit('Initial commit', skip_hooks=True)
        return repo

    @pytest.mark.parametrize('favor_on_conflict', [
        'upstream',
        'downstream',
        None,
    ], ids=[
        'favoring_upstream',
        'favoring_downstream',
        'no_favors',
    ])
    def test__git_rebase(self, rebased_sources, old_sources, new_sources, old_repo, new_repo, favor_on_conflict):
        Patcher.cont = False
        Patcher.non_interactive = True
        Patcher.kwargs = dict(rebased_sources_dir=rebased_sources)
        Patcher.old_sources = old_sources
        Patcher.new_sources = new_sources
        Patcher.old_repo = old_repo
        Patcher.new_repo = new_repo
        Patcher.favor_on_conflict = favor_on_conflict
        Patcher.patches = [PatchObject(os.path.basename(getattr(self, 'PATCH{0}'.format(n))), n, 1)
                           for n in range(1, 5)]
        patches = Patcher._git_rebase()  # pylint: disable=protected-access
        assert patches['untouched'] == [os.path.basename(self.PATCH1)]
        if favor_on_conflict == 'upstream':
            assert patches['modified'] == [os.path.basename(self.PATCH2)]
            assert patches['deleted'] == [os.path.basename(self.PATCH3), os.path.basename(self.PATCH4)]
            assert 'inapplicable' not in patches
        elif favor_on_conflict == 'downstream':
            assert patches['modified'] == [os.path.basename(self.PATCH2), os.path.basename(self.PATCH3)]
            assert patches['deleted'] == [os.path.basename(self.PATCH4)]
            assert 'inapplicable' not in patches
        else:
            assert patches['modified'] == [os.path.basename(self.PATCH2)]
            assert patches['deleted'] == [os.path.basename(self.PATCH4)]
            assert patches['inapplicable'] == [os.path.basename(self.PATCH3)]
        with open(os.path.join(rebased_sources, os.path.basename(self.PATCH2))) as f:
            content = f.read()
            assert 'From: {0} <{1}>\n'.format(self.USER, self.EMAIL) in content
            assert 'Subject: [PATCH] P2\n' in content
            assert Patcher.decorate_patch_name(os.path.basename(self.PATCH2)) not in content



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os
import json

from typing import Any, Dict, Tuple

from rebasehelper.exceptions import CheckerNotFoundError, RebaseHelperError
from rebasehelper.helpers.rpm_helper import RpmHelper
from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.plugins.checkers import BaseChecker


class Rpminspect(BaseChecker):  # pylint: disable=abstract-method
    """An abstract class representing rpminspect checker (either RPM or SRPM).

    For both types of checkers the data stored in results_store
    looks as follows:

    {
        'checks': {
            'package1': {
                'BAD': 5,
                'VERIFY': 3,
                'OK': 1,
                'INFO': 1
            },
            'package2': {
                'BAD': 1,
                'VERIFY': 3,
                'OK': 5,
                'INFO': 1
            }
        },
        'path': rebase-helper-results/checkers/rpminspect,
        'files': [
            'package1.json',
            'package2.json'
        ]
    }
    """
    DEFAULT: bool = True

    CMD: str = 'rpminspect'

    BAD_CHECK: str = 'BAD'
    VERIFY_CHECK: str = 'VERIFY'
    OK_CHECK: str = 'OK'
    INFO_CHECK: str = 'INFO'
    POSSIBLE_CHECK_RESULTS = (BAD_CHECK, VERIFY_CHECK, OK_CHECK, INFO_CHECK)

    @classmethod
    def is_available(cls):
        try:
            return ProcessHelper.run_subprocess([cls.CMD, '--help'], output_file=ProcessHelper.DEV_NULL) == 0
        except (IOError, OSError):
            return False

    @classmethod
    def format(cls, data):
        output_lines = [cls.get_underlined_title(cls.name)]

        for package, package_data in data['checks'].items():
            output_lines.append(' - {}:'.format(package))
            for check in cls.POSSIBLE_CHECK_RESULTS:
                output_lines.append('\t- {} {} checks'.format(package_data[check], check))

        output_lines.append('Details in {}:'.format(data['path']))
        for file in data['files']:
            output_lines.append(' - {}'.format(file))
        return output_lines

    @classmethod
    def get_important_changes(cls, checker_output):
        if any(data[cls.BAD_CHECK] > 0 for data in checker_output['checks'].values()):
            return ['{} checks found in {} output. Check rpminspect output.'.format(cls.BAD_CHECK, cls.name)]
        return []

    @classmethod
    def process_data(cls, data: Dict[str, Any]) -> Dict[str, int]:
        result = {check: 0 for check in cls.POSSIBLE_CHECK_RESULTS}
        for check_data in data.values():
            for check_dict in check_data:
                status = check_dict['result']
                result[status] += 1
        return result

    @classmethod
    def run_rpminspect(cls, checker_dir: str, old_pkg: str, new_pkg: str) -> Tuple[str, Dict[str, int]]:
        """Runs rpminspect on the given package.

        Args:
            checker_dir: Path to the results directory of the checker
            old_pkg: Path to the old package.
            new_pkg: Path to the new package.

        Returns:
            Tuple of (path, data), where path is the path to rpminspect output
            and data is the dict of count of OK/BAD/VERIFY checks.

        """
        cmd = [cls.CMD, '-F', 'json']
        pkg_name = RpmHelper.split_nevra(os.path.basename(new_pkg))['name']
        cmd.append(old_pkg)
        cmd.append(new_pkg)

        outfile = os.path.join(checker_dir, '{}.json'.format(pkg_name))
        try:
            ret = ProcessHelper.run_subprocess(cmd, output_file=outfile, ignore_stderr=True)
        except OSError as e:
            raise CheckerNotFoundError('Checker \'{}\' was not found or installed.'.format(cls.name)) from e

        # Exit code 1 is used when bad check is found, 0 when everything is OK. Others on error
        if ret not in (0, 1):
            raise RebaseHelperError('An error occurred when running checker \'{}\''.format(cls.name))

        with open(outfile, 'r') as json_file:
            data = json.load(json_file)

        return outfile, cls.process_data(data)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os

import pytest  # type: ignore

from rebasehelper.constants import RESULTS_DIR, DEBUG_LOG, CHANGES_PATCH, REPORT


def make_artifacts_report():
    artifacts = [
        DEBUG_LOG,
        CHANGES_PATCH,
        REPORT + '.json',
        'old-build/SRPM/build.log',
        'old-build/RPM/build.log',
        'old-build/RPM/root.log',
        'old-build/RPM/mock_output.log',
        'new-build/SRPM/build.log',
        'new-build/RPM/build.log',
        'new-build/RPM/root.log',
        'new-build/RPM/mock_output.log',
    ]
    report = []
    for artifact in artifacts:
        try:
            with open(os.path.join(RESULTS_DIR, artifact)) as f:
                content = f.read()
                report.append(' {} '.format(artifact).center(80, '_'))
                report.append(content)
        except IOError:
            continue
    return '\n'.join(report)


@pytest.mark.hookwrapper
def pytest_runtest_makereport():
    outcome = yield
    report = outcome.get_result()
    report.sections.append(('Artifacts', make_artifacts_report()))



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.plugin import Plugin
from rebasehelper.plugins.plugin_collection import PluginCollection
from rebasehelper.results_store import results_store
from rebasehelper.constants import RESULTS_DIR, REPORT, LOGS_DIR, DEBUG_LOG, OLD_BUILD_DIR, NEW_BUILD_DIR


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))
logger_summary: CustomLogger = cast(CustomLogger, logging.getLogger('rebasehelper.summary'))


class BaseOutputTool(Plugin):
    """Base class for an output tool"""

    DEFAULT: bool = False
    EXTENSION: str = ''

    @classmethod
    def get_report_path(cls, app):
        return os.path.join(app.results_dir, REPORT + '.' + cls.EXTENSION)

    @classmethod
    def prepend_results_dir_name(cls, app, *path_members):
        """Prepends a path with path to rebase-helper-results.

        Takes directory changes (such as cd into cloned repo with --bugzilla-id)
        into account.

        Args:
            app: Application instance.
            path_members: Path members to prepend the results dir to.

        Returns:
            Prepended path.

        """
        if app.conf.results_dir:
            if os.path.isabs(app.conf.results_dir):
                prepend = app.conf.results_dir
            else:
                # make the path relative to start_dir
                prepend = os.path.relpath(os.path.join(app.execution_dir, app.conf.results_dir), app.start_dir)
            path = os.path.join(prepend, RESULTS_DIR, *path_members)
        else:
            path = os.path.join(os.path.relpath(os.getcwd(), app.start_dir), RESULTS_DIR, *path_members)
        if not path.startswith(os.pardir) and not path.startswith(os.curdir):
            path = os.path.join(os.curdir, path)
        return path

    @classmethod
    def print_cli_summary(cls, app):
        """Outputs a summary of a rebase-helper run.

        Args:
            app: Application instance.
        """
        cls.print_patches_cli()
        result = results_store.get_result_message()

        cls.print_important_checkers_output()

        logger_summary.heading('\nAvailable logs:')
        logger_summary.info('%s:\n%s', 'Debug log', cls.prepend_results_dir_name(app,
                                                                                 os.path.join(LOGS_DIR, DEBUG_LOG)))
        if results_store.get_old_build() is not None:
            logger_summary.info('%s:\n%s', 'Old build logs and (S)RPMs',
                                cls.prepend_results_dir_name(app, OLD_BUILD_DIR))
        if results_store.get_new_build() is not None:
            logger_summary.info('%s:\n%s', 'New build logs and (S)RPMs',
                                cls.prepend_results_dir_name(app, NEW_BUILD_DIR))
        logger_summary.info('')

        logger_summary.heading('%s:', 'Rebased sources')
        logger_summary.info("%s", cls.prepend_results_dir_name(app, os.path.relpath(app.rebased_sources_dir,
                                                                                    app.results_dir)))

        patch = results_store.get_changes_patch()
        if 'changes_patch' in patch:
            logger_summary.heading('%s:', 'Generated patch')
            logger_summary.info("%s\n", cls.prepend_results_dir_name(app, os.path.basename(patch['changes_patch'])))

        cls.print_report_file_path(app)

        if 'success' in result:
            logger_summary.success('\n%s', result['success'])
        # Error is printed out through exception caught in CliHelper.run()

    @classmethod
    def print_important_checkers_output(cls):
        """Iterates over all checkers output to highlight important checkers warning"""
        checkers_results = results_store.get_checkers()
        for check_tool in cls.manager.checkers.plugins.values():
            if check_tool:
                for check, data in sorted(checkers_results.items()):
                    if check == check_tool.name:
                        out = check_tool.get_important_changes(data)
                        if out:
                            logger_summary.warning('\n'.join(out))

    @classmethod
    def print_report_file_path(cls, app):
        """Outputs path to the report file"""
        logger_summary.heading('%s report:', cls.name)
        logger_summary.info('%s', cls.prepend_results_dir_name(app, 'report.' + cls.EXTENSION))

    @classmethod
    def print_patches_cli(cls):
        """Outputs info about patches"""
        patch_dict = {
            'inapplicable': logger_summary.error,
            'modified': logger_summary.success,
            'deleted': logger_summary.success,
        }

        for patch_type, logger_method in patch_dict.items():
            cls.print_patches_section_cli(logger_method, patch_type)

    @classmethod
    def print_patches_section_cli(cls, logger_method, patch_type):
        """Outputs information about one of the patch types.

        Args:
            logger_method: Method to use for logging
            patch_type: A key to use for filtering patch dictionary obtained
                        from results_store.
        """
        patches = results_store.get_patches()
        if not patches:
            return

        if patch_type in patches:
            logger_summary.info('\n%s patches:', patch_type)
            for patch in sorted(patches[patch_type]):
                logger_method(patch)

    @classmethod
    def run(cls, logs, app):  # pylint: disable=unused-argument
        raise NotImplementedError()


class OutputToolCollection(PluginCollection):
    """
    Class representing the process of running various output tools.
    """

    def run(self, tool, logs=None, app=None):
        """Runs specified output tool.

        Args:
            tool(str): Tool to run.
            logs (list): Build logs containing information about the failed rebase.
            app (rebasehelper.application.Application): Application class instance.

        """
        output_tool = self.plugins[tool]
        logger.info("Running '%s' output tool.", tool)
        output_tool.run(logs, app=app)
        output_tool.print_cli_summary(app)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import subprocess
import tempfile
from typing import cast

from rebasehelper.constants import SYSTEM_ENCODING
from rebasehelper.logger import CustomLogger


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class ProcessHelper:

    """Class for executing subprocesses."""

    DEV_NULL: str = os.devnull

    @staticmethod
    def run_subprocess(cmd, input_file=None, output_file=None, ignore_stderr=False):
        """Runs the specified command in a subprocess.

        Args:
            cmd (iterable): A sequence of program arguments.
            input_file (str, typing.BytesIO): File to read the input from.
            output_file (str, typing.BytesIO): File to write the output of the command to.
            ignore_stderr (bool): Whether to ignore stderr output.

        Returns:
            int: Exit code of the subprocess.

        """
        return ProcessHelper.run_subprocess_cwd(cmd,
                                                input_file=input_file,
                                                output_file=output_file,
                                                ignore_stderr=ignore_stderr)

    @staticmethod
    def run_subprocess_cwd(cmd, cwd=None, input_file=None, output_file=None, ignore_stderr=False, shell=False):
        """Runs the specified command in a subprocess in a different working directory.

        Args:
            cmd (iterable): A sequence of program arguments.
            cwd (str): Working directory for the command.
            input_file (str, typing.BytesIO): File to read the input from.
            output_file (str, typing.BytesIO): File to write the output of the command to.
            ignore_stderr (bool): Whether to ignore stderr output.
            shell (bool): Whether to run the command in a shell.

        Returns:
            int: Exit code of the subprocess.

        """
        return ProcessHelper.run_subprocess_cwd_env(cmd,
                                                    cwd=cwd,
                                                    input_file=input_file,
                                                    output_file=output_file,
                                                    ignore_stderr=ignore_stderr,
                                                    shell=shell)

    @staticmethod
    def run_subprocess_env(cmd, env=None, input_file=None, output_file=None, ignore_stderr=False, shell=False):
        """Runs the specified command in a subprocess with a redefined environment.

        Args:
            cmd (iterable): A sequence of program arguments.
            env (dict): Environment variables for the new process.
            input_file (str, typing.BytesIO): File to read the input from.
            output_file (str, typing.BytesIO): File to write the output of the command to.
            ignore_stderr (bool): Whether to ignore stderr output.
            shell (bool): Whether to run the command in a shell.

        Returns:
            int: Exit code of the subprocess.

        """
        return ProcessHelper.run_subprocess_cwd_env(cmd,
                                                    env=env,
                                                    input_file=input_file,
                                                    output_file=output_file,
                                                    ignore_stderr=ignore_stderr,
                                                    shell=shell)

    @staticmethod
    def run_subprocess_cwd_env(cmd, cwd=None, env=None, input_file=None, output_file=None, ignore_stderr=False,
                               shell=False):
        """Runs the specified command in a subprocess in a different working directory with a redefined environment.

        Args:
            cmd (iterable): A sequence of program arguments.
            cwd (str): Working directory for the command.
            env (dict): Environment variables for the new process.
            input_file (str, typing.BytesIO): File to read the input from.
            output_file (str, typing.BytesIO): File to write the output of the command to.
            ignore_stderr (bool): Whether to ignore stderr output.
            shell (bool): Whether to run the command in a shell.

        Returns:
            int: Exit code of the subprocess.

        """
        close_out_file = False
        close_in_file = False

        logger.debug("cmd=%s, cwd=%s, env=%s, input_file=%s, output_file=%s, shell=%s",
                     str(cmd), str(cwd), str(env), str(input_file), str(output_file), str(shell))

        # write the output to a file/file-like object?
        try:
            out_file = open(output_file, 'wb')
        except TypeError:
            out_file = output_file
        else:
            close_out_file = True

        # read the input from a file/file-like object?
        try:
            in_file = open(input_file, 'r')
        except TypeError:
            in_file = input_file
        else:
            close_in_file = True

        # we need to rewind the file object pointer to the beginning
        try:
            in_file.seek(0)
        except AttributeError:
            # we don't mind - in_file might be None
            pass

        # check if in_file has fileno() method - which is needed for Popen
        try:
            in_file.fileno()
        except (AttributeError, OSError):
            spooled_in_file = tempfile.SpooledTemporaryFile(mode='w+b')
            try:
                in_data = in_file.read()
            except AttributeError:
                spooled_in_file.close()
            else:
                spooled_in_file.write(in_data.encode(SYSTEM_ENCODING))
                spooled_in_file.seek(0)
                in_file = spooled_in_file
                close_in_file = True

        # need to change environment variables?
        if env is not None:
            local_env = os.environ.copy()
            local_env.update(env)
        else:
            local_env = None

        if out_file:
            stdout = subprocess.PIPE
        else:
            stdout = None

        with open(os.devnull, 'wb') as devnull:
            sp = subprocess.Popen(cmd,
                                  stdin=in_file,
                                  stdout=stdout,
                                  stderr=devnull if ignore_stderr else subprocess.STDOUT,
                                  cwd=cwd,
                                  env=local_env,
                                  shell=shell)

        if out_file is not None:
            # read the output
            for line in sp.stdout:
                try:
                    out_file.write(line.decode(SYSTEM_ENCODING))
                except TypeError:
                    out_file.write(line)
            # TODO: Need to figure out how to send output to stdout (without logger) and to logger
            # else:
            #   logger.debug(line.rstrip("\n"))

        # we need to rewind the file object pointer to the beginning
        try:
            out_file.seek(0)
        except AttributeError:
            # we don't mind - out_file might be None
            pass

        if close_out_file:
            out_file.close()

        if close_in_file:
            in_file.close()

        sp.wait()

        logger.debug("subprocess exited with return code %s", str(sp.returncode))

        return sp.returncode



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.plugin import Plugin
from rebasehelper.plugins.plugin_collection import PluginCollection
from rebasehelper.types import PackageCategories


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class BaseSpecHook(Plugin):
    """Base class for a spec hook"""

    CATEGORIES: PackageCategories = []

    @classmethod
    def run(cls, spec_file, rebase_spec_file, **kwargs):
        """
        Runs a spec hook.

        :param spec_file: Original spec file object
        :param rebase_spec_file: Rebased spec file object
        :param kwargs: Keyword arguments from Application instance
        """
        raise NotImplementedError()


class SpecHookCollection(PluginCollection):
    """
    Class representing the process of running various spec file hooks.
    """

    def run(self, spec_file, rebase_spec_file, **kwargs):
        """Runs all non-blacklisted spec hooks.

        Args:
            spec_file (rebasehelper.specfile.SpecFile): Original SpecFile object.
            rebase_spec_file (rebasehelper.specfile.SpecFile): Rebased SpecFile object.
            **kwargs: Keyword arguments from Application instance.

        """
        blacklist = kwargs.get("spec_hook_blacklist", [])

        for name, spec_hook in self.plugins.items():
            if not spec_hook or name in blacklist:
                continue
            categories = spec_hook.CATEGORIES
            if not categories or spec_file.category in categories:
                logger.info("Running '%s' spec hook", name)
                spec_hook.run(spec_file, rebase_spec_file, **kwargs)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import collections.abc
import fnmatch
import re
from typing import Dict, Iterator, List, Optional, Tuple, cast

from rebasehelper.spec_content import SpecContent
from rebasehelper.helpers.macro_helper import MacroHelper


class Tag:
    def __init__(self, section_index: int, section_name: str, line: int, name: str, value_span: Tuple[int, int],
                 valid: bool, index: Optional[int] = None) -> None:
        self.section_index: int = section_index
        self.section_name: str = section_name
        self.line: int = line
        self.name: str = name
        self.value_span: Tuple[int, int] = value_span
        self.valid: bool = valid
        self.index: Optional[int] = index

    def __eq__(self, other):
        return (self.section_index == other.section_index and
                self.section_name == other.section_name and
                self.line == other.line and
                self.name == other.name and
                self.value_span == other.value_span and
                self.valid == other.valid and
                self.index == other.index)

    def __str__(self) -> str:
        return (
            "Tag("
            f"section_index='{self.section_index}', "
            f"section_name='{self.section_name}', "
            f"line='{self.line}', "
            f"name='{self.name}', "
            f"value_span='{self.value_span}', "
            f"valid='{self.valid}', "
            f"index='{self.index}')"
        )


class Tags(collections.abc.Sequence):
    def __init__(self, raw_content: SpecContent, parsed_content: SpecContent) -> None:
        self.items: Tuple[Tag] = self._parse(raw_content, parsed_content)

    def __getitem__(self, index):
        return self.items[index]

    def __len__(self):
        return len(self.items)

    @classmethod
    def _map_sections_to_parsed(cls, raw_content: SpecContent, parsed_content: SpecContent) -> Dict[int, int]:
        """Creates a mapping of raw sections to parsed sections.

        Returns:
            A dict where the keys are indexes of sections in raw content and values are their
            counterparts in parsed content. Value is set to -1 if the section is not in
            parsed SpecContent.
        """
        result: Dict[int, int] = {}
        parsed = 0
        for index, (section, _) in enumerate(raw_content.sections):
            if parsed < len(parsed_content.sections) and parsed_content.sections[parsed][0].lower() == section.lower():
                result[index] = parsed
                parsed += 1
            else:
                result[index] = -1

        return result

    @classmethod
    def _parse(cls, raw_content: SpecContent, parsed_content: SpecContent) -> Tuple[Tag]:
        result: List[Tag] = []
        parsed_mapping = cls._map_sections_to_parsed(raw_content, parsed_content)
        next_source_index = 0
        next_patch_index = 0
        # counts number of occurrences of one particular section
        for section_index, (section, section_content) in enumerate(raw_content.sections):
            section = section.lower()
            parsed_section_index = parsed_mapping[section_index]
            parsed: List[str] = [] if parsed_section_index == -1 else parsed_content.sections[parsed_section_index][1]
            if section.startswith('%package'):
                tags, next_source_index, next_patch_index = cls._parse_package_tags(section, section_content, parsed,
                                                                                    section_index, next_source_index,
                                                                                    next_patch_index)
                result.extend(tags)
            elif section.startswith('%sourcelist') or section.startswith('%patchlist'):
                tags, next_source_index, next_patch_index = cls._parse_list_tags(section, section_content, parsed,
                                                                                 section_index, next_source_index,
                                                                                 next_patch_index)
                result.extend(tags)
        return cast(Tuple[Tag], tuple(result))

    @classmethod
    def _sanitize_tag(cls, name: str, next_source_index: int,
                      next_patch_index: int) -> Tuple[str, Optional[int], int, int]:
        """Sanitizes a tag name.

        Capitalizes the tag. If the tag is numbered, strips padded zeroes.
        Implements automatic numbering of Patches and Sources (adds the number
        to their name).
        """
        name, *rest = re.split(r'(\d+)', name.lower(), 1)
        sanitized_name = name.capitalize()
        next_index = next_source_index if name == 'source' else next_patch_index
        index = int(next(iter(rest), next_index))
        if name == 'source':
            return '{}{}'.format(sanitized_name, index), index, index + 1, next_patch_index
        if name == 'patch':
            return '{}{}'.format(sanitized_name, index), index, next_source_index, index + 1
        return sanitized_name, None, next_source_index, next_patch_index

    @classmethod
    def _parse_package_tags(cls, section: str, section_content: List[str], parsed: List[str], section_index: int,
                            next_source_index: int, next_patch_index: int) -> Tuple[List[Tag], int, int]:
        """Parses all tags in a %package section and determines if they are valid.

        A tag is considered valid if it is still present after evaluating all conditions.

        Note that this is not perfect - if the same tag appears in both %if and %else blocks,
        and has the same value in both, it's impossible to tell them apart, so only the latter
        is considered valid, disregarding the actual condition.

        Returns:
              A tuple containing: a tuple of all Tags object, new next source index, new next patch index.

              Indexed tag names are sanitized, for example 'Source' is replaced with 'Source0'
              and 'Patch007' with 'Patch7'.

              Tag names are capitalized, section names are lowercase.

        """
        result = []
        tag_re = re.compile(r'^(?P<prefix>(?P<name>\w+)\s*:\s*)(?P<value>.+)$')
        for line_index, line in enumerate(section_content):
            expanded = MacroHelper.expand(line)
            if not line or not expanded:
                continue
            valid = bool(parsed and [p for p in parsed if p == expanded.rstrip()])
            m = tag_re.match(line)
            if m:
                tag_name, tag_index, next_source_index, next_patch_index = cls._sanitize_tag(m.group('name'),
                                                                                             next_source_index,
                                                                                             next_patch_index)
                result.append(Tag(section_index, section, line_index, tag_name, m.span('value'), valid, tag_index))
                continue
            m = tag_re.match(expanded)
            if m:
                start = line.find(m.group('prefix'))
                if start < 0:
                    # tag is probably defined by a macro, just ignore it
                    continue
                # conditionalized tag
                line = line[start:].rstrip('}')  # FIXME: removing trailing braces is not very robust
                m = tag_re.match(line)
                if m:
                    span = cast(Tuple[int, int], tuple(x + start for x in m.span('value')))
                    tag_name, tag_index, next_source_index, next_patch_index = cls._sanitize_tag(m.group('name'),
                                                                                                 next_source_index,
                                                                                                 next_patch_index)
                    result.append(Tag(section_index, section, line_index, tag_name, span, valid, tag_index))

        return result, next_source_index, next_patch_index

    @classmethod
    def _parse_list_tags(cls, section: str, section_content: List[str], parsed: List[str], section_index: int,
                         next_source_index: int, next_patch_index: int) -> Tuple[List[Tag], int, int]:
        """Parses all tags in a %sourcelist or %patchlist section.

        Only parses tags that are valid (that is - are in parsed), nothing more can
        consistently be detected.

        Follows how rpm works, the new Source/Patch tags are indexed starting from
        the last parsed Source/Patch tag.

        """
        tag = 'Source' if section == '%sourcelist' else 'Patch'
        result = []
        for i, line in enumerate(section_content):
            expanded = MacroHelper.expand(line)
            is_comment = SpecContent.get_comment_span(line, section)[0] != len(line)
            if not expanded or not line or is_comment or not [p for p in parsed if p == expanded.rstrip()]:
                continue
            tag_name, tag_index, next_source_index, next_patch_index = cls._sanitize_tag(tag, next_source_index,
                                                                                         next_patch_index)
            result.append(Tag(section_index, section, i, tag_name, (0, len(line)), True, tag_index))

        return result, next_source_index, next_patch_index

    @classmethod
    def _filter(cls, tags: Tuple[Tag], section_index: Optional[int] = None, section_name: Optional[str] = None,
                name: Optional[str] = None, valid: Optional[bool] = True) -> Iterator[Tag]:
        result = iter(tags)
        if section_index is not None:
            result = filter(lambda t: t.section_index == section_index, result)
        if section_name is not None:
            result = filter(lambda t: t.section_name == section_name.lower(), result)  # type: ignore
        if name is not None:
            result = filter(lambda t: fnmatch.fnmatchcase(t.name, name.capitalize()), result)  # type: ignore
        if valid is not None:
            result = filter(lambda t: t.valid == valid, result)
        return result

    def filter(self, section_index: Optional[int] = None, section_name: Optional[str] = None,
               name: Optional[str] = None, valid: Optional[bool] = True) -> Iterator[Tag]:
        """Filters tags based on section, name or validity. Defaults to all valid tags in all sections.

        Args:
            section_index: If specified, includes tags only from section of this index.
            section_name: If specified, includes tags only from sections of this name.
            name: If specified, includes tags matching this name. Wildcards are supported.
            valid: If specified, includes tags of this validity.

        Returns:
            Iterator of matching Tag objects.

        """
        return self._filter(self.items, section_index, section_name, name, valid)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import copy
import os
import shutil

import pytest  # type: ignore
import rpm  # type: ignore

from rebasehelper.specfile import SpecFile
from rebasehelper.spec_content import SpecContent
from rebasehelper.tags import Tags
from rebasehelper.helpers.macro_helper import MacroHelper


TESTS_DIR: str = os.path.dirname(__file__)
TEST_FILES_DIR: str = os.path.join(TESTS_DIR, 'testing_files')
SPEC_FILE: str = 'test.spec'


@pytest.yield_fixture(autouse=True)
def workdir(request, tmpdir_factory):
    with tmpdir_factory.mktemp('workdir').as_cwd():
        wd = os.getcwd()
        # copy testing files into workdir
        for file_name in getattr(request.cls, 'TEST_FILES', []):
            shutil.copy(os.path.join(TEST_FILES_DIR, file_name), wd)
        yield wd


@pytest.fixture
def spec_object(workdir):  # pylint: disable=redefined-outer-name
    shutil.copy(os.path.join(TEST_FILES_DIR, SPEC_FILE), workdir)
    return SpecFile(SPEC_FILE, workdir)


@pytest.fixture
def mocked_spec_object(spec_attributes):
    spec = SpecFile.__new__(SpecFile)
    spec.save = lambda: None
    for attribute, value in spec_attributes.items():
        if attribute == 'macros':
            for macro, properties in value.items():
                rpm.addMacro(macro, properties.get('value', ''))
            macros = MacroHelper.dump()
            for macro, properties in value.items():
                for m in macros:
                    if m['name'] == macro:
                        for prop, v in properties.items():
                            if prop != 'value':
                                m[prop] = v
            value = macros
        if attribute == 'spec_content' and isinstance(value, str):
            value = SpecContent(value)
        setattr(spec, attribute, value)
    if hasattr(spec, 'spec_content') and not hasattr(spec, 'tags'):
        spec.tags = Tags(spec.spec_content, spec.spec_content)
    return spec


@pytest.fixture
def mocked_spec_object_copy(mocked_spec_object):  # pylint: disable=redefined-outer-name
    return copy.deepcopy(mocked_spec_object)


def pytest_collection_modifyitems(items):
    for item in items:
        # item is an instance of Function class.
        # https://github.com/pytest-dev/pytest/blob/master/_pytest/python.py
        if 'functional' in item.fspath.strpath:
            item.add_marker(pytest.mark.functional)
        else:
            item.add_marker(pytest.mark.standard)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.output_tools import BaseOutputTool
from rebasehelper.plugins.checkers import CheckerCategory
from rebasehelper.logger import LoggerHelper
from rebasehelper.results_store import results_store
from rebasehelper.constants import RESULTS_DIR


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))
logger_report: CustomLogger = cast(CustomLogger, logging.getLogger('rebasehelper.report'))


class Text(BaseOutputTool):

    """ Text output tool. """

    DEFAULT: bool = True
    EXTENSION: str = 'txt'

    @classmethod
    def print_success_message(cls, results):
        """Outputs a success/failure message depending on the result of a rebase.

        Args:
            results: ResultsStore instance to get data from.

        """
        results = results.get_result_message()
        if 'success' in results:
            logger_report.info(results['success'])
        else:
            logger_report.info(results['fail'])

    @classmethod
    def print_changes_patch(cls, results):
        """Outputs the location of changes.patch.

        Args:
            results: ResultsStore instance to get data from.

        """
        patch = results.get_changes_patch()
        if patch is not None:
            logger_report.info('\nPatch with differences between old and new version source files:')
            logger_report.info(os.path.join(RESULTS_DIR, os.path.basename(patch['changes_patch'])))

    @classmethod
    def print_message_and_separator(cls, message="", separator='='):
        logger_report.info(message)
        logger_report.info(separator * (len(message) - 1))

    @classmethod
    def print_patches(cls, patches):
        cls.print_message_and_separator("\nDownstream Patches")
        if not patches:
            logger_report.info("Patches were neither modified nor deleted.")
            return

        logger_report.info("Rebased patches are located in %s", os.path.join(RESULTS_DIR, 'rebased-sources'))
        logger_report.info("Legend:")
        logger_report.info("[-] = already applied, patch removed")
        logger_report.info("[*] = merged, patch modified")
        logger_report.info("[!] = conflicting or inapplicable, patch skipped")
        logger_report.info("[ ] = patch untouched")

        patches_out = list()
        for patch_type, patch_list in sorted(patches.items()):
            if patch_list:
                symbols = dict(deleted='-', modified='*', inapplicable='!')
                for patch in sorted(patch_list):
                    patches_out.append(' * {0:40} [{1}]'.format(os.path.basename(patch),
                                                                symbols.get(patch_type, ' ')))
        logger_report.info('\n'.join(sorted(patches_out)))

    @classmethod
    def print_rpms_and_logs(cls, rpms, version):
        """Outputs information about location of RPMs and logs created during rebase.

        Args:
            rpms: Dictionary of (S)RPM paths.
            version: "new" or "old" string.

        """
        pkgs = ['srpm', 'rpm']
        if not rpms.get('rpm', None):
            return
        message = '\n{} packages'.format(version)
        cls.print_message_and_separator(message=message, separator='-')
        for type_rpm in pkgs:
            srpm = rpms.get(type_rpm, None)
            if not srpm:
                continue

            if type_rpm == 'srpm':
                message = "\nSource packages and logs are in directory %s:"
            else:
                message = "\nBinary packages and logs are in directory %s:"

            if isinstance(srpm, str):
                # Print SRPM path
                dirname = os.path.dirname(srpm)
                logger_report.info(message, os.path.join(RESULTS_DIR, version.lower() + '-build', 'SRPM'))
                logger_report.info(" - %s", os.path.basename(srpm))
                # Print SRPM logs
                cls.print_build_logs(rpms, dirname)

            else:
                # Print RPMs paths
                dirname = os.path.dirname(srpm[0])
                logger_report.info(message, os.path.join(RESULTS_DIR, version.lower() + '-build', 'RPM'))
                for pkg in sorted(srpm):
                    logger_report.info(" - %s", os.path.basename(pkg))
                # Print RPMs logs
                cls.print_build_logs(rpms, dirname)

    @classmethod
    def print_build_logs(cls, rpms, dirpath):
        """Outputs paths to rpm build logs.

        Args:
            rpms: Dictionary of (S)RPM paths.
            dirpath: Directory of the currently processed (S)RPM.

        """
        if rpms.get('logs', None) is None:
            return
        for logs in sorted(rpms.get('logs', []) + rpms.get('srpm_logs', [])):
            if dirpath not in logs:
                # Skip logs that do not belong to curent rpms(and version)
                continue
            logger_report.info(' - %s', os.path.basename(logs))

    @classmethod
    def print_summary(cls, report_path, results):
        """Outputs a summary of a rebase.

        Args:
            report_path: Where the report file should be created
            results: ResultsStore instance to get data from.

        """
        info = results.get_summary_info()
        if info:
            for key, value in info.items():
                level = logging.getLevelName(key)
                if isinstance(level, int):
                    logger.log(level, value)
                else:
                    logger.info('%s: %s', key, value)

        LoggerHelper.add_file_handler(logger_report, report_path)

        cls.print_success_message(results)
        logger_report.info("All result files are stored in %s", os.path.dirname(report_path))

        cls.print_changes_patch(results)

        cls.print_checkers_text_output(results.get_checkers())

        cls.print_build_log_hooks_result(results.get_build_log_hooks())

        if results.get_patches():
            cls.print_patches(results.get_patches())

        cls.print_message_and_separator("\nRPMS")
        for pkg_version in ['old', 'new']:
            pkg_results = results.get_build(pkg_version)
            if pkg_results:
                cls.print_rpms_and_logs(pkg_results, pkg_version.capitalize())

    @classmethod
    def print_checkers_text_output(cls, checkers_results):
        """Outputs text output of every checker."""
        for category in (CheckerCategory.SOURCE, CheckerCategory.SRPM, CheckerCategory.RPM):
            for check_tool in cls.manager.checkers.plugins.values():
                if check_tool and check_tool.get_category() == category:
                    for check, data in sorted(checkers_results.items()):
                        if check == check_tool.name:
                            logger_report.info('\n'.join(check_tool.format(data)))

    @classmethod
    def print_build_log_hooks_result(cls, build_log_hooks_result):
        for hook, data in build_log_hooks_result.items():
            if data:
                cls.print_message_and_separator('\n{} build log hook'.format(hook))
                logger_report.info('\n'.join(cls.manager.build_log_hooks.get_plugin(hook).format(data)))

    @classmethod
    def run(cls, logs, app):  # pylint: disable=unused-argument
        cls.print_summary(cls.get_report_path(app), results_store)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.build_tools import RpmbuildTemporaryEnvironment
from rebasehelper.plugins.build_tools.srpm import SRPMBuildToolBase
from rebasehelper.helpers.path_helper import PathHelper
from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.exceptions import SourcePackageBuildError


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class Rpmbuild(SRPMBuildToolBase):

    DEFAULT: bool = True

    CMD: str = 'rpmbuild'

    @classmethod
    def _build_srpm(cls, spec, workdir, results_dir, srpm_results_dir, srpm_builder_options):
        """Builds SRPM using rpmbuild.

        Args:
            spec: Path to SPEC file to build the SRPM from.
            workdir: Path to working directory with rpmbuild directory structure.
            results_dir: Path to directory where logs will be placed.
            srpm_results_dir: Path to directory where SRPM will be placed.
            srpm_builder_options: Additional rpmbuild options.

        Returns:
            Tuple, the first element is path to SRPM, the second is a list of paths
            to logs.

        """
        logger.info("Building SRPM")
        spec_loc, spec_name = os.path.split(spec)
        output = os.path.join(results_dir, "build.log")

        cmd = [cls.CMD, '-bs', spec_name]

        if srpm_builder_options is not None:
            cmd.extend(srpm_builder_options)

        ret = ProcessHelper.run_subprocess_cwd_env(cmd,
                                                   cwd=spec_loc,
                                                   env={'HOME': workdir},
                                                   output_file=output)

        build_log_path = os.path.join(srpm_results_dir, 'build.log')
        logs = []
        for l in PathHelper.find_all_files(results_dir, '*.log'):
            logs.append(os.path.join(srpm_results_dir, os.path.basename(l)))

        if ret == 0:
            return PathHelper.find_first_file(workdir, '*.src.rpm'), logs
        # An error occurred, raise an exception
        raise SourcePackageBuildError("Building SRPM failed!", logfile=build_log_path, logs=logs)

    @classmethod
    def build(cls, spec, results_dir, **kwargs):
        """
        Build SRPM with chosen SRPM Build Tool

        :param spec: SpecFile object
        :param results_dir: absolute path to DIR where results should be stored
        :return: absolute path to SRPM, list with absolute paths to logs
        """
        sources = spec.get_sources()
        patches = [p.path for p in spec.get_patches()]
        with RpmbuildTemporaryEnvironment(sources, patches, spec.path, results_dir) as tmp_env:
            srpm_builder_options = cls.get_srpm_builder_options(**kwargs)

            env = tmp_env.env()
            tmp_dir = tmp_env.path()
            tmp_spec = env.get(RpmbuildTemporaryEnvironment.TEMPDIR_SPEC)
            tmp_results_dir = env.get(
                RpmbuildTemporaryEnvironment.TEMPDIR_RESULTS)

            srpm, logs = cls._build_srpm(tmp_spec, tmp_dir, tmp_results_dir, results_dir,
                                         srpm_builder_options=srpm_builder_options)

        logger.info("Building SRPM finished successfully")

        # srpm path in results_dir
        srpm = os.path.join(results_dir, os.path.basename(srpm))
        logger.verbose("Successfully built SRPM: '%s'", str(srpm))
        logger.verbose("logs: '%s'", str(logs))

        return dict(srpm=srpm, logs=logs)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

from typing import Dict, List, Optional, Type, Union, TYPE_CHECKING

from rebasehelper.plugins.plugin_loader import PluginLoader
from rebasehelper.plugins.plugin import Plugin
from rebasehelper.types import Options

if TYPE_CHECKING:
    # avoid cyclic import at runtime
    from rebasehelper.plugins.plugin_manager import PluginManager


class PluginCollection:
    def __init__(self, entrypoint: str, manager: 'PluginManager'):
        self.plugins: Dict[str, Optional[Type[Plugin]]] = PluginLoader.load(entrypoint, manager)

    def get_all_plugins(self) -> List[str]:
        return list(self.plugins)

    def get_supported_plugins(self) -> List[str]:
        return [k for k, v in self.plugins.items() if v]

    def get_default_plugins(self, return_one: bool = False) -> Union[str, List[str], None]:
        default = [k for k, v in self.plugins.items() if v and getattr(v, 'DEFAULT', False)]
        return default if not return_one else default[0] if default else None

    def get_plugin(self, tool: str) -> Optional[Type[Plugin]]:
        try:
            return self.plugins[tool]
        except KeyError as e:
            raise NotImplementedError("Unsupported plugin") from e

    def get_options(self) -> Options:
        """Gets options of all plugins of one type.

        Returns:
            list: List of plugins' options.

        """
        options: List = []
        for plugin in self.plugins.values():
            if plugin:
                options.extend(plugin.OPTIONS)

        return options



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
from typing import cast

from rebasehelper.helpers.process_helper import ProcessHelper
from rebasehelper.helpers.input_helper import InputHelper
from rebasehelper.helpers.path_helper import PathHelper
from rebasehelper.helpers.rpm_helper import RpmHelper
from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.build_tools import RpmbuildTemporaryEnvironment
from rebasehelper.plugins.build_tools.rpm import BuildToolBase
from rebasehelper.exceptions import RebaseHelperError, BinaryPackageBuildError


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class Rpmbuild(BuildToolBase):  # pylint: disable=abstract-method
    """
    Class representing rpmbuild build tool.
    """

    ACCEPTS_OPTIONS: bool = True

    CMD: str = 'rpmbuild'

    @classmethod
    def _build_rpm(cls, srpm, workdir, results_dir, rpm_results_dir, builder_options=None):
        """Builds RPMs using rpmbuild

        Args:
            srpm: Path to SRPM.
            workdir: Path to working directory with rpmbuild directory structure.
            results_dir: Path to directory where logs will be placed.
            rpm_results_dir: Path to directory where RPMs will be placed.
            builder_options: Additional options for rpmbuild.

        Returns:
            Tuple, the first element is a list of paths to built RPMs,
            the second is a list of paths to logs.

        """
        logger.info("Building RPMs")
        output = os.path.join(results_dir, "build.log")

        cmd = [cls.CMD, '--rebuild', srpm]
        if builder_options is not None:
            cmd.extend(builder_options)
        ret = ProcessHelper.run_subprocess_cwd_env(cmd,
                                                   env={'HOME': workdir},
                                                   output_file=output)

        build_log_path = os.path.join(rpm_results_dir, 'build.log')
        logs = []
        for log in PathHelper.find_all_files(results_dir, '*.log'):
            logs.append(os.path.join(rpm_results_dir, os.path.basename(log)))

        if ret == 0:
            return [f for f in PathHelper.find_all_files(workdir, '*.rpm') if not f.endswith('.src.rpm')], logs
        # An error occurred, raise an exception
        raise BinaryPackageBuildError("Building RPMs failed!", results_dir, logfile=build_log_path, logs=logs)

    @classmethod
    def prepare(cls, spec, conf):
        """
        Checks if all build dependencies are installed. If not, asks user whether they should be installed.
        If he agrees, installs build dependencies using PolicyKit.

        :param spec: SpecFile object
        """
        if not RpmHelper.all_packages_installed(spec.header.requires):
            question = '\nSome build dependencies are missing. Do you want to install them now'
            if conf.non_interactive or InputHelper.get_message(question):
                if RpmHelper.install_build_dependencies(spec.path, assume_yes=conf.non_interactive) != 0:
                    raise RebaseHelperError('Failed to install build dependencies')

    @classmethod
    def build(cls, spec, results_dir, srpm, **kwargs):
        """
        Builds the RPMs using rpmbuild

        :param spec: SpecFile object
        :param results_dir: absolute path to DIR where results should be stored
        :param srpm: absolute path to SRPM
        :return: dict with:
                 'rpm' -> list with absolute paths to RPMs
                 'logs' -> list with absolute paths to build_logs
        """
        sources = spec.get_sources()
        patches = [p.path for p in spec.get_patches()]
        with RpmbuildTemporaryEnvironment(sources, patches, spec.path, results_dir) as tmp_env:
            env = tmp_env.env()
            tmp_dir = tmp_env.path()
            tmp_results_dir = env.get(RpmbuildTemporaryEnvironment.TEMPDIR_RESULTS)
            rpms, logs = cls._build_rpm(srpm, tmp_dir, tmp_results_dir, results_dir,
                                        builder_options=cls.get_builder_options(**kwargs))

        logger.info("Building RPMs finished successfully")

        # RPMs paths in results_dir
        rpms = [os.path.join(results_dir, os.path.basename(f)) for f in rpms]
        logger.verbose("Successfully built RPMs: '%s'", str(rpms))
        logger.verbose("logs: '%s'", str(logs))

        return dict(rpm=rpms, logs=logs)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import collections
import logging
import os
import re
from typing import Dict, List, Optional, Set, Union, cast

from rebasehelper.logger import CustomLogger
from rebasehelper.results_store import results_store
from rebasehelper.plugins.checkers import BaseChecker, CheckerCategory
from rebasehelper.helpers.rpm_helper import RpmHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))

SonameChanges = Dict[str, Dict[str, List[Union[str, Dict[str, str]]]]]


class SonameCheck(BaseChecker):

    DEFAULT: bool = True
    CATEGORY: Optional[CheckerCategory] = CheckerCategory.RPM

    @classmethod
    def is_available(cls):
        return True

    @classmethod
    def _get_sonames(cls, provides: List[str]) -> Set[str]:
        soname_re = re.compile(r'(?P<soname>[^()]+\.so[^()]+)\(.*\)(\(64bit\))?')
        sonames = set()
        for p in provides:
            match = soname_re.match(p)
            if match:
                sonames.add(match.group('soname'))
        return sonames

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        old_headers = [RpmHelper.get_header_from_rpm(x) for x in results_store.get_old_build().get('rpm', [])]
        new_headers = [RpmHelper.get_header_from_rpm(x) for x in results_store.get_new_build().get('rpm', [])]
        soname_changes: SonameChanges = collections.defaultdict(lambda: collections.defaultdict(list))
        for old in old_headers:
            new = [x for x in new_headers if x.name == old.name]
            if not new:
                logger.warning('New version of package %s was not found!', old.name)
                continue
            else:
                new = new[0]
            old_sonames = cls._get_sonames(old.provides)
            new_sonames = cls._get_sonames(new.provides)
            for old_soname in old_sonames:
                if old_soname in new_sonames:
                    new_sonames.remove(old_soname)
                    continue
                soname = [x for x in new_sonames if os.path.splitext(x)[0] == os.path.splitext(old_soname)[0]]
                if not soname:
                    soname_changes[old.name]['removed'].append(old_soname)
                else:
                    soname_changes[old.name]['changed'].append({'from': old_soname, 'to': soname[0]})
                    new_sonames.remove(soname[0])

            if new_sonames:
                soname_changes[old.name]['added'] = list(new_sonames)

        return dict(soname_changes=soname_changes)

    @classmethod
    def format(cls, data):
        output_lines = [cls.get_underlined_title('sonamecheck')]
        if not data['soname_changes']:
            output_lines.append('No SONAME changes occurred.')
            return output_lines
        for package, package_changes in data['soname_changes'].items():
            output_lines.append(package)
            for change_type, changes in package_changes.items():
                output_lines.append(' - {}'.format(change_type))
                for change in changes:
                    if change_type == 'changed':
                        output_lines.append('\t- from {} to {}'.format(change['from'], change['to']))
                    else:
                        output_lines.append('\t - {}'.format(change))

        return output_lines

    @classmethod
    def get_important_changes(cls, checker_output):
        if checker_output['soname_changes']:
            return ['SONAME changes occurred. Check sonamecheck output in the report.']
        return []



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

from rebasehelper.plugins.spec_hooks import BaseSpecHook
from rebasehelper.helpers.macro_helper import MacroHelper


class PathsToRPMMacros(BaseSpecHook):
    """SpecHook for replacing paths to files with RPM macros."""

    @classmethod
    def run(cls, spec_file, rebase_spec_file, **kwargs):
        macros = [m for m in rebase_spec_file.macros if m['name'] in MacroHelper.MACROS_WHITELIST]
        macros = MacroHelper.expand_macros(macros)
        # ensure maximal greediness
        macros.sort(key=lambda k: len(k['value']), reverse=True)

        for sec_name, sec_content in rebase_spec_file.spec_content.sections:
            if sec_name.startswith('%files'):
                for index, line in enumerate(sec_content):
                    new_path = MacroHelper.substitute_path_with_macros(line, macros)
                    sec_content[index] = new_path
        rebase_spec_file.save()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import sys
import time
import urllib.error
import urllib.request
from typing import cast

import requests

from rebasehelper.exceptions import DownloadError
from rebasehelper.logger import CustomLogger


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class DownloadHelper:

    """Class for downloading files and performing HTTP requests."""

    @staticmethod
    def progress(download_total, downloaded, start_time):
        """Prints current progress and estimated remaining time of a download to the standard output.

        Args:
            download_total (int): Total download size in bytes.
            downloaded (int): Size of the already downloaded portion of a file in bytes.
            start_time (float): Time when the download started in seconds since epoch.

        """
        bar_width = 32
        infinite_step = 256 * 1024  # move every 256 kilobytes

        delta = time.time() - start_time

        def format_time(t):
            h, rem = divmod(int(t), 3600)
            m, s = divmod(rem, 60)
            return '{:0>2d}:{:0>2d}:{:0>2d}'.format(h, m, s)

        def format_size(s):
            units = [' ', 'K', 'M', 'G', 'T']
            i = 0
            while s >= 1024.0 and i < len(units) - 1:
                s /= 1024.0
                i += 1
            return '{:>7.2F}{}'.format(s, units[i])

        if download_total < 0:
            # infinite progress bar
            pct = ' ' * 4
            pos = int(downloaded / infinite_step) % (bar_width - 5)
            bar = '[{}<=>{}]'.format(' ' * pos, ' ' * (bar_width - 5 - pos))
            ts = ' in {}'.format(format_time(delta))
        else:
            r = float(downloaded) / float(download_total) if download_total else 0.0
            pct = '{:>3d}%'.format(int(r * 100))
            pos = int(r * (bar_width - 3))
            bar = '[{}>{}]'.format('=' * pos, ' ' * (bar_width - 3 - pos))
            ts = 'eta {}'.format(format_time(delta / r - delta) if r > 0.0 else ' ' * 7 + '?')

        size = format_size(downloaded)

        # no point to log progress, write directly to stdout
        sys.stdout.write('\r{}{}  {}  {} '.format(pct, bar, size, ts))
        sys.stdout.flush()

    @staticmethod
    def request(url, **kwargs):
        """Performs an HTTP request or an FTP RETR command.

        Args:
            url (str): HTTP, HTTPS or FTP URL.
            **kwargs: Keyword arguments to be passed to requests.session.get().

        Returns:
            requests.Response: Response object.

        """

        class FTPAdapter(requests.adapters.BaseAdapter):

            def send(self, request, stream=False, timeout=None, verify=True,  # pylint: disable=unused-argument
                     cert=None, proxies=None):  # pylint: disable=unused-argument
                response = requests.models.Response()
                response.request = request
                response.connection = self
                try:
                    resp = urllib.request.urlopen(request.url)
                except urllib.error.URLError as e:
                    response.status_code = 400
                    response.reason = e.reason
                else:
                    response.status_code = 200
                    response.headers = requests.structures.CaseInsensitiveDict(getattr(resp, 'headers', {}))
                    response.raw = resp
                    response.url = resp.url
                return response

            def close(self):
                pass

        session = requests.Session()
        session.mount('ftp://', FTPAdapter())

        try:
            return session.get(url, **kwargs)
        except requests.exceptions.RequestException as e:
            logger.error('%s: %s', type(e).__name__, str(e))
            return None

    @staticmethod
    def download_file(url, destination_path, blocksize=8192):
        """Downloads a file from HTTP, HTTPS or FTP URL.

        Args:
            url (str): URL to be downloaded.
            destination_path (str): Path to where the downloaded file will be stored.
            blocksize (int): Block size in bytes.

        """
        r = DownloadHelper.request(url, stream=True)
        if r is None:
            raise DownloadError("An unexpected error occurred during the download.")

        if not 200 <= r.status_code < 300:
            raise DownloadError(r.reason)

        file_size = int(r.headers.get('content-length', -1))
        if r.headers.get('content-encoding', 'identity') != 'identity':
            # use infinite progress bar in case of compressed content, there is no
            # way to determine the uncompressed size in advance
            file_size = -1

        # file exists, check the size
        if os.path.exists(destination_path):
            if file_size < 0 or file_size != os.path.getsize(destination_path):
                logger.verbose("The destination file '%s' exists, but sizes don't match! Removing it.",
                               destination_path)
                os.remove(destination_path)
            else:
                logger.verbose("The destination file '%s' exists, and the size is correct! Skipping download.",
                               destination_path)
                return
        try:
            with open(destination_path, 'wb') as local_file:
                logger.info('Downloading file from URL %s', url)
                download_start = time.time()
                downloaded = 0

                # report progress
                DownloadHelper.progress(file_size, downloaded, download_start)

                # do the actual download
                for chunk in r.iter_content(chunk_size=blocksize):
                    downloaded += len(chunk)
                    local_file.write(chunk)

                    # report progress
                    DownloadHelper.progress(file_size, downloaded, download_start)

                sys.stdout.write('\n')
                sys.stdout.flush()
        except KeyboardInterrupt as e:
            os.remove(destination_path)
            raise e



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os

from rebasehelper.types import Options
from rebasehelper.constants import CONFIG_PATH, CONFIG_FILENAME, CHANGES_PATCH
from rebasehelper.plugins.plugin_manager import plugin_manager


OPTIONS: Options = [
    # basic
    {
        "name": ["--version"],
        "default": False,
        "switch": True,
        "help": "show rebase-helper version and exit",
    },
    # output control
    {
        "name": ["-v", "--verbose"],
        "default": 0,
        "counter": True,
        "help": "be more verbose",
    },
    {
        "name": ["--color"],
        "choices": ["always", "never", "auto"],
        "default": "auto",
        "help": "colorize the output, defaults to %(default)s",
    },
    {
        "name": ["--background"],
        "choices": ["dark", "light", "auto"],
        "default": "auto",
        "help": "use color scheme for the given background, defaults to %(default)s",
    },
    {
        "name": ["--results-dir"],
        "help": "location where the rebase-helper-results directory will be created",
    },
    {
        "name": ["--workspace-dir"],
        "help": "location where the rebase-helper-workspace directory will be created",
    },
    # tool selection
    {
        "name": ["--buildtool"],
        "choices": plugin_manager.build_tools.get_all_plugins(),
        "available_choices": plugin_manager.build_tools.get_supported_plugins(),
        "default": plugin_manager.build_tools.get_default_plugins(True),
        "help": "build tool to use, defaults to %(default)s",
    },
    {
        "name": ["--srpm-buildtool"],
        "choices": plugin_manager.srpm_build_tools.get_all_plugins(),
        "available_choices": plugin_manager.srpm_build_tools.get_supported_plugins(),
        "default": plugin_manager.srpm_build_tools.get_default_plugins(True),
        "help": "SRPM build tool to use, defaults to %(default)s",
    },
    {
        "name": ["--pkgcomparetool"],
        "choices": plugin_manager.checkers.get_all_plugins(),
        "available_choices": plugin_manager.checkers.get_supported_plugins(),
        "nargs": "?",
        "const": [""],
        "default": plugin_manager.checkers.get_default_plugins(),
        "type": lambda s: s.split(','),
        "help": "set of tools to use for package comparison, defaults to "
                "%(default)s if available",
    },
    {
        "name": ["--outputtool"],
        "choices": plugin_manager.output_tools.get_all_plugins(),
        "available_choices": plugin_manager.output_tools.get_supported_plugins(),
        "default": plugin_manager.output_tools.get_default_plugins(True),
        "help": "tool to use for formatting rebase output, defaults to %(default)s",
    },
    {
        "name": ["--versioneer"],
        "choices": plugin_manager.versioneers.get_all_plugins(),
        "available_choices": plugin_manager.versioneers.get_supported_plugins(),
        "default": None,
        "help": "tool to use for determining latest upstream version",
    },
    # blacklists
    {
        "name": ["--versioneer-blacklist"],
        "choices": plugin_manager.versioneers.get_all_plugins(),
        "available_choices": plugin_manager.versioneers.get_supported_plugins(),
        "nargs": "?",
        "const": [""],
        "default": [],
        "type": lambda s: s.split(","),
        "help": "prevent specified versioneers from being run",
    },
    {
        "name": ["--spec-hook-blacklist"],
        "choices": plugin_manager.spec_hooks.get_all_plugins(),
        "available_choices": plugin_manager.spec_hooks.get_supported_plugins(),
        "nargs": "?",
        "const": [""],
        "default": [],
        "type": lambda s: s.split(","),
        "help": "prevent specified spec hooks from being run",
    },
    {
        "name": ["--build-log-hook-blacklist"],
        "choices": plugin_manager.build_log_hooks.get_all_plugins(),
        "available_choices": plugin_manager.build_log_hooks.get_supported_plugins(),
        "nargs": "?",
        "const": [""],
        "default": [],
        "type": lambda s: s.split(","),
        "help": "prevent specified build log hooks from being run"
    },
    # behavior control
    {
        "name": ["--bugzilla-id"],
        "metavar": "BUG_ID",
        "default": None,
        "help": "do a rebase based on Upstream Release Monitoring bugzilla"
    },
    {
        "name": ["--non-interactive"],
        "default": False,
        "switch": True,
        "dest": "non_interactive",
        "help": "do not interact with user",
    },
    {
        "name": ["--favor-on-conflict"],
        "choices": ["downstream", "upstream", "off"],
        "default": "off",
        "dest": "favor_on_conflict",
        "help": "favor downstream or upstream changes when conflicts appear",
    },
    {
        "name": ["--not-download-sources"],
        "default": False,
        "switch": True,
        "help": "do not download sources",
    },
    {
        "name": ["-w", "--keep-workspace"],
        "default": False,
        "switch": True,
        "help": "do not remove workspace directory after finishing",
    },
    {
        "name": ["--apply-changes"],
        "default": False,
        "switch": True,
        "help": "apply {} after a successful rebase".format(CHANGES_PATCH),
    },
    {
        "name": ["--disable-inapplicable-patches"],
        "default": False,
        "switch": True,
        "dest": "disable_inapplicable_patches",
        "help": "disable inapplicable patches in rebased SPEC file",
    },
    {
        "name": ["--skip-version-check"],
        "default": False,
        "switch": True,
        "help": "force rebase even if current version is newer than requested version",
    },
    {
        "name": ["--update-sources"],
        "default": False,
        "switch": True,
        "help": "update \"sources\" file and upload new sources to lookaside cache",
    },
    {
        "name": ["--skip-upload"],
        "default": False,
        "switch": True,
        "help": "skip uploading new sources to lookaside cache",
    },
    {
        "name": ["--force-build-log-hooks"],
        "default": False,
        "switch": True,
        "help": "enforce running of build log hooks (even in non-interactive mode)",
    },
    # remote builder options
    {
        "name": ["--builds-nowait"],
        "default": False,
        "switch": True,
        "help": "do not wait for remote builds to finish",
    },
    {
        "name": ["--build-tasks"],
        "dest": "build_tasks",
        "metavar": "OLD_TASK,NEW_TASK",
        "type": lambda s: s.split(','),
        "help": "comma-separated remote build task ids",
    },
    # additional local builder options
    {
        "name": ["--builder-options"],
        "default": None,
        "metavar": "BUILDER_OPTIONS",
        "help": "enable arbitrary local builder option(s), enclose %(metavar)s in quotes "
                "to pass more than one",
    },
    {
        "name": ["--srpm-builder-options"],
        "default": None,
        "metavar": "SRPM_BUILDER_OPTIONS",
        "help": "enable arbitrary local srpm builder option(s), enclose %(metavar)s in quotes "
                "to pass more than one",
    },
    # misc
    {
        "name": ["--changelog-entry"],
        "default": "- New upstream release %{version}",
        "help": "text to use as changelog entry, can contain RPM macros, which will be expanded",
    },
    {
        "name": ["--config-file"],
        "default": os.path.join(CONFIG_PATH, CONFIG_FILENAME),
        "help": "path to a configuration file, defaults to %(default)s",
    },
    {
        "name": ["-D", "--define"],
        "default": [],
        "append": True,
        "dest": "rpmmacros",
        "metavar": "'MACRO EXPR'",
        "help": "define an rpm macro, can be used multiple times",
    },
    # sources
    {
        "name": ["sources"],
        "metavar": "SOURCES",
        "nargs": "?",
        "default": None,
        "help": "version number or filename of the new source archive",
    },
]


def traverse_options(options):
    group_index = 0
    for opt in options:
        if isinstance(opt, list):
            for inner_opt in opt:
                yield dict(group=group_index, **inner_opt)
            group_index += 1
        else:
            yield opt



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import re
from typing import List

import rpm  # type: ignore
from pkg_resources import parse_version

from rebasehelper.helpers.console_helper import ConsoleHelper


class MacroHelper:

    """Class for working with RPM macros"""

    MACROS_WHITELIST: List[str] = [
        '_bindir',
        '_datadir',
        '_includedir',
        '_infodir',
        '_initdir',
        '_libdir',
        '_libexecdir',
        '_localstatedir',
        '_mandir',
        '_sbindir',
        '_sharedstatedir',
        '_sysconfdir',
        'python2_sitelib',
        'python3_sitelib',
    ]

    @staticmethod
    def expand(s, default=None):
        try:
            return rpm.expandMacro(s)
        except rpm.error:
            return default

    @classmethod
    def purge_macro(cls, macro: str) -> None:
        m = '%{{{}}}'.format(macro)
        while cls.expand(m, m) != m:
            rpm.delMacro(macro)

    @classmethod
    def expand_macros(cls, macros):
        """Expands values of multiple macros.

        Args:
            macros (list): List of macros to be expanded, macros are
                represented as dicts.

        Returns:
            list: List of macros with expanded values.

        """
        for macro in macros:
            macro['value'] = cls.expand(macro['value'])
        return macros

    @staticmethod
    def substitute_path_with_macros(path, macros):
        """Substitutes parts of a path with macros.

        Args:
            path (str): Path to be changed.
            macros (list): Macros which can be used as a substitution.

        Returns:
            str: Path expressed using macros.

        """
        for m in macros:
            if m['value'] and m['value'] in path:
                path = path.replace(m['value'], '%{{{}}}'.format(m['name']))

        return path

    @staticmethod
    def dump():
        """Gets list of all defined macros.

        Returns:
            list: All defined macros.

        """
        macro_re = re.compile(
            r'''
            ^\s*
            (?P<level>-?\d+)
            (?P<used>=|:)
            [ ]
            (?P<name>\w+)
            (?P<options>\(.+?\))?
            [\t]
            (?P<value>.*)
            $
            ''',
            re.VERBOSE)

        with ConsoleHelper.Capturer(stderr=True) as capturer:
            rpm.expandMacro('%dump')

        macros = []

        def add_macro(properties):
            macro = dict(properties)
            macro['used'] = macro['used'] == '='
            macro['level'] = int(macro['level'])
            if parse_version(rpm.__version__) < parse_version('4.13.90'):
                # in RPM < 4.13.90 level of some macros is decreased by 1
                if macro['level'] == -1:
                    # this could be macro with level -1 or level 0, we can not be sure
                    # so just duplicate the macro for both levels
                    macros.append(macro)
                    macro = dict(macro)
                    macro['level'] = 0
                    macros.append(macro)
                elif macro['level'] in (-14, -16):
                    macro['level'] += 1
                    macros.append(macro)
                else:
                    macros.append(macro)
            else:
                macros.append(macro)

        for line in capturer.stderr.split('\n'):
            match = macro_re.match(line)
            if match:
                add_macro(match.groupdict())

        return macros

    @staticmethod
    def filter(macros, **kwargs):
        """Finds all macros satisfying certain conditions.

        Args:
            macros (list): Macros to be filtered.
            **kwargs: Filters to be used.

        Returns:
            list: Macros satisfying the conditions.

        """
        def _test(macro):
            return all(macro.get(k[4:]) >= v if k.startswith('min_') else
                       macro.get(k[4:]) <= v if k.startswith('max_') else
                       macro.get(k) == v for k, v in kwargs.items())

        return [m for m in macros if _test(m)]



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import copy

from typing import Dict, Any


class ResultsStore:
    """Class for storing information about results from rebase-helper actions."""

    RESULTS_INFORMATION: str = 'information'
    RESULTS_CHECKERS: str = 'checkers'
    RESULTS_BUILD_LOG_HOOKS: str = 'build_log_hooks'
    RESULTS_BUILDS: str = 'builds'
    RESULTS_PATCHES: str = 'patches'
    RESULTS_CHANGES_PATCH: str = 'changes_patch'
    RESULTS_SUCCESS: str = 'result'

    def __init__(self):
        self._data_store: Dict[str, Any] = dict()

    def clear(self):
        self._data_store.clear()

    def set_results(self, results_type, data_dict):
        if results_type not in (
                self.RESULTS_INFORMATION,
                self.RESULTS_CHECKERS,
                self.RESULTS_BUILD_LOG_HOOKS,
                self.RESULTS_BUILDS,
                self.RESULTS_PATCHES,
                self.RESULTS_CHANGES_PATCH,
                self.RESULTS_SUCCESS
        ):
            raise ValueError('Trying to set unsupported type of results: {}!'.format(results_type))

        try:
            dict_to_update = self._data_store[results_type]
        except KeyError:
            dict_to_update = dict()
            self._data_store[results_type] = dict_to_update
        dict_to_update.update(data_dict)

    def set_info_text(self, text, data):
        self.set_results(self.RESULTS_INFORMATION, {text: data})

    def set_patches_results(self, results_dict):
        self.set_results(self.RESULTS_PATCHES, results_dict)

    def set_checker_output(self, text, data):
        self.set_results(self.RESULTS_CHECKERS, {text: data})

    def set_build_log_hooks_result(self, text, data):
        self.set_results(self.RESULTS_BUILD_LOG_HOOKS, {text: data})

    def set_build_data(self, version, data):
        self.set_results(self.RESULTS_BUILDS, {version: data})

    def set_changes_patch(self, text, data):
        self.set_results(self.RESULTS_CHANGES_PATCH, {text: data})

    def set_result_message(self, text, data):
        self.set_results(self.RESULTS_SUCCESS, {text: data})

    def get_all(self):
        return copy.deepcopy(self._data_store)

    def get_build(self, version):
        builds_results = self._data_store.get(self.RESULTS_BUILDS, None)
        if builds_results is not None:
            return builds_results.get(version, None)
        else:
            return None

    def get_old_build(self):
        return self.get_build('old')

    def get_new_build(self):
        return self.get_build('new')

    def get_patches(self):
        return self._data_store.get(self.RESULTS_PATCHES, None)

    def get_checkers(self):
        return self._data_store.get(self.RESULTS_CHECKERS, {})

    def get_build_log_hooks(self):
        return self._data_store.get(self.RESULTS_BUILD_LOG_HOOKS, {})

    def get_summary_info(self):
        return self._data_store.get(self.RESULTS_INFORMATION, None)

    def get_changes_patch(self):
        return self._data_store.get(self.RESULTS_CHANGES_PATCH, None)

    def get_result_message(self):
        return self._data_store.get(self.RESULTS_SUCCESS, None)


# global results store
results_store: ResultsStore = ResultsStore()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import io

import pytest  # type: ignore

from rebasehelper.helpers.input_helper import InputHelper


class TestInputHelper:

    @pytest.mark.parametrize('suffix, answer, kwargs, expected_input', [
        (' [Y/n]? ', 'yes', None, True),
        (' [Y/n]? ', 'no', None, False),
        (' [y/N]? ', 'yes', dict(default_yes=False), True),
        (' [Y/n]? ', '\n', None, True),
        (' [y/N]? ', '\n', dict(default_yes=False), False),
        (' ', 'random input\ndsfdf', dict(any_input=True), True),
        (' ', 'random input\n', dict(default_yes=False, any_input=True), False),
    ], ids=[
        'yes',
        'no',
        'yes-default_no',
        'no_input-default_yes',
        'no_input-default_no',
        'any_input-default_yes',
        'any_input-default_no',
    ])
    def test_get_message(self, monkeypatch, capsys, suffix, answer, kwargs, expected_input):
        question = 'bla bla'
        monkeypatch.setattr('sys.stdin', io.StringIO(answer))
        inp = InputHelper.get_message(question, **(kwargs or {}))
        assert capsys.readouterr()[0] == question + suffix
        assert inp is expected_input



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import pytest  # type: ignore

from rebasehelper.helpers.rpm_helper import RpmHelper


class TestRpmHelper:
    """ RpmHelper class tests. """

    def test_is_package_installed_existing(self):
        assert RpmHelper.is_package_installed('glibc') is True
        assert RpmHelper.is_package_installed('coreutils') is True

    def test_is_package_installed_non_existing(self):
        assert RpmHelper.is_package_installed('non-existing-package') is False
        assert RpmHelper.is_package_installed('another-non-existing-package') is False

    def test_all_packages_installed_existing(self):
        assert RpmHelper.all_packages_installed(['glibc', 'coreutils']) is True

    def test_all_packages_installed_one_non_existing(self):
        assert RpmHelper.all_packages_installed(['glibc', 'coreutils', 'non-existing-package']) is False

    @pytest.mark.parametrize('nevra, name, epoch, version, release, arch', [
        ('libtiff-static-4.0.7-5.fc26.x86_64', 'libtiff-static', None, '4.0.7', '5.fc26', 'x86_64'),
        ('libtiff-debuginfo-4.0.8-1.fc25', 'libtiff-debuginfo', None, '4.0.8', '1.fc25', None),
        ('libtiff-tools.i686', 'libtiff-tools', None, None, None, 'i686'),
        ('libtiff-devel', 'libtiff-devel', None, None, None, None),
        ('libpng-devel-debuginfo-2:1.6.31-1.fc27.aarch64', 'libpng-devel-debuginfo', 2, '1.6.31', '1.fc27', 'aarch64'),
        (
                'python-genmsg-0.3.10-14.20130617git95ca00d.fc28',
                'python-genmsg',
                None,
                '0.3.10',
                '14.20130617git95ca00d.fc28',
                None,
        ),
        (
                'golang-github-MakeNowJust-heredoc-devel-0-0.9.gitbb23615.fc28.noarch',
                'golang-github-MakeNowJust-heredoc-devel',
                None,
                '0',
                '0.9.gitbb23615.fc28',
                'noarch',
        ),
    ], ids=[
        'libtiff-static-4.0.7-5.fc26.x86_64',
        'libtiff-debuginfo-4.0.8-1.fc25',
        'libtiff-tools.i686',
        'libtiff-devel',
        'libpng-devel-debuginfo-2:1.6.31-1.fc27.aarch64',
        'python-genmsg-0.3.10-14.20130617git95ca00d.fc28',
        'golang-github-MakeNowJust-heredoc-devel-0-0.9.gitbb23615.fc28.noarch',
    ])
    def test_split_nevra(self, nevra, name, epoch, version, release, arch):
        assert RpmHelper.split_nevra(nevra) == dict(name=name, epoch=epoch, version=version,
                                                    release=release, arch=arch)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os

import pytest  # type: ignore

from typing import List

from rebasehelper.archive import Archive


class TestArchive:
    TAR_GZ: str = 'archive.tar.gz'
    TGZ: str = 'archive.tgz'
    TAR_XZ: str = 'archive.tar.xz'
    TAR_BZ2: str = 'archive.tar.bz2'
    ZIP: str = 'archive.zip'
    BZ2: str = 'file.txt.bz2'
    CRATE: str = 'archive.crate'
    GEM: str = 'archive.gem'
    INVALID_TAR_BZ2: str = 'archive-invalid.tar.bz2'
    INVALID_TAR_XZ: str = 'archive-invalid.tar.xz'

    ARCHIVED_FILE: str = 'file.txt'
    ARCHIVED_FILE_CONTENT: str = 'simple testing file'

    TEST_FILES: List[str] = [os.path.join('archives', archive) for archive in [
        TAR_GZ,
        TGZ,
        TAR_XZ,
        TAR_BZ2,
        ZIP,
        BZ2,
        CRATE,
        GEM,
        INVALID_TAR_BZ2,
        INVALID_TAR_XZ,
    ]]

    @pytest.fixture
    def extracted_archive(self, archive, workdir):
        a = Archive(archive)
        d = os.path.join(workdir, 'dir')
        a.extract_archive(d)
        if archive == self.GEM:
            # append top level dir
            d = os.path.join(d, 'archive')
        return d

    @pytest.mark.parametrize('archive', [
        TAR_GZ,
        TGZ,
        TAR_XZ,
        TAR_BZ2,
        BZ2,
        ZIP,
        CRATE,
        GEM,
    ], ids=[
        'tar.gz',
        'tgz',
        'tar.xz',
        'tar.bz2',
        'bz2',
        'zip',
        'crate',
        'gem',
    ])
    def test_archive(self, extracted_archive):
        extracted_file = os.path.join(extracted_archive, self.ARCHIVED_FILE)
        #  check if the dir was created
        assert os.path.isdir(extracted_archive)
        #  check if the file was extracted
        assert os.path.isfile(extracted_file)
        #  check the content
        with open(extracted_file) as f:
            assert f.read().strip() == self.ARCHIVED_FILE_CONTENT

    @pytest.mark.parametrize('archive', [
        INVALID_TAR_BZ2,
        INVALID_TAR_XZ,
    ], ids=[
        'tar.bz2',
        'tar.xz',
    ])
    def test_invalid_archive(self, archive, workdir):
        a = Archive(archive)
        d = os.path.join(workdir, 'dir')
        with pytest.raises(IOError):
            a.extract_archive(d)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os

import pytest  # type: ignore

from typing import List

from rebasehelper.cli import CLI
from rebasehelper.config import Config
from rebasehelper.application import Application
from rebasehelper import constants


class TestApplication:
    OLD_SOURCES: str = 'test-1.0.2.tar.xz'
    NEW_SOURCES: str = 'test-1.0.3.tar.xz'
    SPEC_FILE: str = 'test.spec'
    PATCH_1: str = 'test-testing.patch'
    PATCH_2: str = 'test-testing2.patch'
    PATCH_3: str = 'test-testing3.patch'
    SOURCE_2: str = 'documentation.tar.xz'
    SOURCE_3: str = 'misc.zip'
    SOURCE_TESTS: str = 'source-tests.sh'

    TEST_FILES: List[str] = [
        OLD_SOURCES,
        NEW_SOURCES,
        SPEC_FILE,
        PATCH_1,
        PATCH_2,
        PATCH_3,
        SOURCE_2,
        SOURCE_3,
        SOURCE_TESTS,
        'rebase-helper-d70cb5a2f523db5b6088427563531f43b7703859.tar.gz',
    ]

    cmd_line_args: List[str] = ['--not-download-sources', '1.0.3']

    @pytest.fixture
    def make_config(self):
        def wrapper(args):
            cli = CLI(args)
            config = Config()
            config.merge(cli)
            return config
        return wrapper

    @pytest.fixture
    def app(self, make_config):
        config = make_config(self.cmd_line_args)
        execution_dir, results_dir = Application.setup(config)
        app = Application(config, os.getcwd(), execution_dir, results_dir)
        return app

    def test_application_sources(self, workdir, app):
        expected_dict = {
            'new': {
                'sources': [os.path.join(workdir, self.SOURCE_TESTS),
                            os.path.join(workdir, self.NEW_SOURCES)],
                'version': '1.0.3',
                'name': 'test',
                'tarball': self.NEW_SOURCES,
                'spec': os.path.join(workdir, constants.RESULTS_DIR, self.SPEC_FILE),
                'patches_full': {1: [os.path.join(workdir, self.PATCH_1),
                                     '',
                                     0,
                                     False],
                                 2: [os.path.join(workdir, self.PATCH_2),
                                     '-p1',
                                     1,
                                     False],
                                 3: [os.path.join(workdir, self.PATCH_3),
                                     '-p1',
                                     2,
                                     False]}},
            'workspace_dir': os.path.join(workdir, constants.WORKSPACE_DIR),
            'old': {
                'sources': [os.path.join(workdir, self.SOURCE_TESTS),
                            os.path.join(workdir, self.OLD_SOURCES)],
                'version': '1.0.2',
                'name': 'test',
                'tarball': self.OLD_SOURCES,
                'spec': os.path.join(workdir, self.SPEC_FILE),
                'patches_full': {1: [os.path.join(workdir, self.PATCH_1),
                                     '',
                                     0,
                                     False],
                                 2: [os.path.join(workdir, self.PATCH_2),
                                     '-p1',
                                     1,
                                     False],
                                 3: [os.path.join(workdir, self.PATCH_3),
                                     '-p1',
                                     2,
                                     False]}},
            'results_dir': os.path.join(workdir, constants.RESULTS_DIR)}

        app.prepare_sources()
        for key, val in app.kwargs.items():
            if key in expected_dict:
                assert val == expected_dict[key]

    def test_setup(self, make_config):
        config = make_config(self.cmd_line_args)
        files = Application.setup(config)
        for file in files:
            assert os.path.exists(file)

    @pytest.mark.parametrize('gitignore, sources, result', [
        (
                [
                    '/source1-*.tar.gz\n',
                    '/source2-2.8.0.tar.xz\n',
                    '/Source3.bz2\n',
                ],
                [
                    'source1-1.0.1.tar.gz',
                    'source2-2.8.1.tar.xz',
                    'Source3.bz2',
                ],
                [
                    '/source1-*.tar.gz\n',
                    '/source2-2.8.0.tar.xz\n',
                    '/Source3.bz2\n',
                    '/source2-2.8.1.tar.xz\n',
                ],
        ),
    ], ids=[
        'gitignore',
    ])
    def test_update_gitignore(self, workdir, gitignore, sources, result):
        with open(os.path.join(workdir, '.gitignore'), 'w') as f:
            for line in gitignore:
                f.write(line)
        Application._update_gitignore(sources, workdir)  # pylint: disable=protected-access
        with open(os.path.join(workdir, '.gitignore')) as f:
            assert f.readlines() == result



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import argparse
import sys

from rebasehelper.exceptions import RebaseHelperError, ParseError


class SilentArgumentParser(argparse.ArgumentParser):
    def __init__(self):
        super(SilentArgumentParser, self).__init__(add_help=False)

    def error(self, message):
        raise ParseError(message)


class CustomHelpFormatter(argparse.HelpFormatter):

    def _expand_help(self, action):
        action.default = getattr(action, 'actual_default', None)
        if isinstance(action.default, list):
            default_str = ','.join(str(c) for c in action.default)
            action.default = default_str
        return super(CustomHelpFormatter, self)._expand_help(action)


class CustomAction(argparse.Action):
    def __init__(self, option_strings,
                 switch=False,
                 counter=False,
                 append=False,
                 actual_default=None,
                 dest=None,
                 default=None,
                 nargs=None,
                 const=None,
                 required=False,
                 type=None,  # pylint: disable=redefined-builtin
                 metavar=None,
                 help=None,  # pylint: disable=redefined-builtin
                 choices=None):

        super(CustomAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            const=const,
            default=default,
            required=required,
            metavar=metavar,
            type=type,
            help=help,
            choices=choices)

        self.switch = switch
        self.counter = counter
        self.append = append
        self.nargs = 0 if self.switch or self.counter else 1 if self.append else nargs
        self.actual_default = actual_default

    def __call__(self, parser, namespace, values, option_string=None):
        if self.counter:
            value = getattr(namespace, self.dest, 0) + 1
        elif self.append:
            value = getattr(namespace, self.dest, [])
            value.extend(values)
        elif self.switch:
            value = True
        else:
            value = values
        setattr(namespace, self.dest, value)


class CustomArgumentParser(argparse.ArgumentParser):

    def _check_value(self, action, value):
        if isinstance(value, list):
            # converted value must be subset of the choices (if specified)
            empty = value == action.const
            if action.choices is not None and not empty and not set(value).issubset(action.choices):
                invalid = set(value).difference(action.choices)
                if len(invalid) == 1:
                    tup = repr(invalid.pop()), ', '.join(map(repr, action.choices))
                    msg = 'invalid choice: %s (choose from %s)' % tup
                else:
                    tup = ', '.join(map(repr, invalid)), ', '.join(map(repr, action.choices))
                    msg = 'invalid choices: %s (choose from %s)' % tup
                raise argparse.ArgumentError(action, msg)
        else:
            super(CustomArgumentParser, self)._check_value(action, value)

    def error(self, message):
        self.print_usage(sys.stderr)
        raise RebaseHelperError(message)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os
import tempfile

from rebasehelper.temporary_environment import TemporaryEnvironment


class TestTemporaryEnvironment:
    """ TemporaryEnvironment class tests. """

    def test_with_statement(self):
        with TemporaryEnvironment() as temp:
            path = temp.path()
            assert path != ''
            assert os.path.exists(path)
            assert os.path.isdir(path)
            env = temp.env()
            assert env.get(temp.TEMPDIR, None) is not None
            assert env.get(temp.TEMPDIR, None) == path

        assert not os.path.exists(path)
        assert not os.path.isdir(path)

    def test_with_statement_exception(self):
        path = ''

        try:
            with TemporaryEnvironment() as temp:
                path = temp.path()
                raise RuntimeError()
        except RuntimeError:
            pass

        assert not os.path.exists(path)
        assert not os.path.isdir(path)

    def test_with_statement_callback(self):
        tmp_file, tmp_path = tempfile.mkstemp(text=True)
        os.close(tmp_file)

        def callback(**kwargs):
            path = kwargs.get(TemporaryEnvironment.TEMPDIR, '')
            assert path != ''
            with open(tmp_path, 'w') as f:
                f.write(path)

        with TemporaryEnvironment(exit_callback=callback) as temp:
            path = temp.path()

        with open(tmp_path, 'r') as f:
            assert f.read() == path

        os.unlink(tmp_path)

    def test_with_statement_callback_exception(self):
        path = ''
        tmp_file, tmp_path = tempfile.mkstemp(text=True)
        os.close(tmp_file)

        def callback(**kwargs):
            path = kwargs.get(TemporaryEnvironment.TEMPDIR, '')
            assert path != ''
            with open(tmp_path, 'w') as f:
                f.write(path)

        try:
            with TemporaryEnvironment(exit_callback=callback) as temp:
                path = temp.path()
                raise RuntimeError()
        except RuntimeError:
            pass

        with open(tmp_path, 'r') as f:
            assert f.read() == path

        os.unlink(tmp_path)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import os
import shutil
from typing import List, Optional, cast

import git  # type: ignore

from rebasehelper.specfile import PatchObject
from rebasehelper.helpers.git_helper import GitHelper
from rebasehelper.helpers.input_helper import InputHelper
from rebasehelper.constants import SYSTEM_ENCODING
from rebasehelper.logger import CustomLogger


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class Patcher:

    """Class for git command used for patching old and new sources"""

    old_sources: Optional[str] = None
    new_sources: Optional[str] = None
    output_data: Optional[str] = None
    old_repo: Optional[git.Repo] = None
    new_repo: Optional[git.Repo] = None
    non_interactive: bool = False
    patches: List[PatchObject] = []

    @staticmethod
    def decorate_patch_name(patch_name):
        return '<<[{0}]>>'.format(patch_name)

    @classmethod
    def insert_patch_name(cls, message, patch_name):
        return '{0}\n\n{1}'.format(message, cls.decorate_patch_name(patch_name))

    @classmethod
    def extract_patch_name(cls, message):
        for line in message.split('\n'):
            if line.startswith('<<[') and line.endswith(']>>'):
                return line[3:-3]
        return None

    @classmethod
    def strip_patch_name(cls, diff, patch_name):
        token = '\n\n{0}'.format(cls.decorate_patch_name(patch_name)).encode(SYSTEM_ENCODING)
        try:
            idx = diff.index(token)
            return diff[:idx] + diff[idx + len(token):]
        except (IndexError, ValueError):
            return diff

    @classmethod
    def apply_patch(cls, repo, patch_object):
        """
        Function applies patches to old sources
        It tries apply patch with am command and if it fails
        then with command --apply
        """
        logger.verbose('Applying patch with git-am')

        patch_name = patch_object.path
        patch_strip = patch_object.strip
        try:
            repo.git.am(patch_name)
            commit = repo.head.commit
        except git.GitCommandError as e:
            logger.verbose('Applying patch with git-am failed.')
            logger.debug(str(e))
            try:
                repo.git.am(abort=True)
            except git.GitCommandError:
                pass
            logger.verbose('Applying patch with git-apply')
            try:
                repo.git.apply(patch_name, p=patch_strip)
            except git.GitCommandError:
                try:
                    repo.git.apply(patch_name, p=patch_strip, reject=True, whitespace='fix')
                except git.GitCommandError as e:
                    logger.verbose('Applying patch with git-apply failed.')
                    logger.debug(str(e))
                    raise
            repo.git.add(all=True)
            commit = repo.index.commit(cls.decorate_patch_name(os.path.basename(patch_name)), skip_hooks=True)
        repo.git.commit(amend=True, m=cls.insert_patch_name(commit.message, os.path.basename(patch_name)))

    @classmethod
    def _git_rebase(cls):
        """Function performs git rebase between old and new sources"""
        def compare_commits(a, b):
            # compare commit diffs disregarding differences in blob hashes
            attributes = (
                'a_mode', 'b_mode', 'a_rawpath', 'b_rawpath',
                'new_file', 'deleted_file', 'raw_rename_from', 'raw_rename_to',
                'diff', 'change_type', 'score')
            diff1 = a.diff(a.parents[0], create_patch=True)
            diff2 = b.diff(b.parents[0], create_patch=True)
            if len(diff1) != len(diff2):
                return False
            for d1, d2 in zip(diff1, diff2):
                for attr in attributes:
                    if getattr(d1, attr) != getattr(d2, attr):
                        return False
            return True
        # in old_sources do:
        # 1) git remote add new_sources <path_to_new_sources>
        # 2) git fetch new_sources
        # 3) git rebase --onto new_sources/master <root_commit_old_sources> <last_commit_old_sources>
        if not os.path.exists(os.path.join(cls.old_sources, '.git', 'rebase-apply')):
            logger.info('git-rebase operation to %s is ongoing...', os.path.basename(cls.new_sources))
            upstream = 'new_upstream'
            cls.old_repo.create_remote(upstream, url=cls.new_sources).fetch()
            root_commit = cls.old_repo.git.rev_list('HEAD', max_parents=0)
            last_commit = cls.old_repo.commit('HEAD')
            if cls.favor_on_conflict == 'upstream':
                strategy_option = 'ours'
            elif cls.favor_on_conflict == 'downstream':
                strategy_option = 'theirs'
            else:
                strategy_option = False
            try:
                cls.output_data = cls.old_repo.git.rebase(root_commit, last_commit,
                                                          strategy_option=strategy_option,
                                                          onto='{}/master'.format(upstream),
                                                          stdout_as_string=True)
            except git.GitCommandError as e:
                ret_code = e.status
                cls.output_data = e.stdout
            else:
                ret_code = 0
        else:
            logger.info('git-rebase operation continues...')
            try:
                cls.output_data = cls.old_repo.git.rebase('--continue', stdout_as_string=True)
            except git.GitCommandError as e:
                ret_code = e.status
                cls.output_data = e.stdout
            else:
                ret_code = 0
        logger.verbose(cls.output_data)
        patch_dictionary = {}
        modified_patches = []
        inapplicable_patches = []
        while ret_code != 0:
            if not cls.old_repo.index.unmerged_blobs() and not cls.old_repo.index.diff(cls.old_repo.commit()):
                # empty commit - conflict has been automatically resolved - skip
                try:
                    cls.output_data = cls.old_repo.git.rebase(skip=True, stdout_as_string=True)
                except git.GitCommandError as e:
                    ret_code = e.status
                    cls.output_data = e.stdout
                    continue
                else:
                    break
            try:
                if os.path.isdir(os.path.join(cls.old_sources, '.git', 'rebase-merge')):
                    with open(os.path.join(cls.old_sources, '.git', 'rebase-merge', 'msgnum')) as f:
                        next_index = int(f.readline())
                    with open(os.path.join(cls.old_sources, '.git', 'rebase-merge', 'end')) as f:
                        last_index = int(f.readline())
                else:
                    with open(os.path.join(cls.old_sources, '.git', 'rebase-apply', 'next')) as f:
                        next_index = int(f.readline())
                    with open(os.path.join(cls.old_sources, '.git', 'rebase-apply', 'last')) as f:
                        last_index = int(f.readline())
            except (FileNotFoundError, IOError) as e:
                raise RuntimeError('Git rebase failed with unknown reason. Please check log file') from e
            patch_name = cls.patches[next_index - 1].get_patch_name()
            inapplicable = False
            if cls.non_interactive:
                inapplicable = True
            else:
                logger.info('Failed to auto-merge patch %s', patch_name)
                unmerged = cls.old_repo.index.unmerged_blobs()
                GitHelper.run_mergetool(cls.old_repo)
                if cls.old_repo.index.unmerged_blobs():
                    if InputHelper.get_message('There are still unmerged entries. Do you want to skip this patch',
                                               default_yes=False):
                        inapplicable = True
                    else:
                        continue
                if not inapplicable:
                    # check for unresolved conflicts
                    unresolved = []
                    for file in unmerged:
                        try:
                            with open(os.path.join(cls.old_sources, file), 'rb') as f:
                                if [l for l in f if b'<<<<<<<' in l]:
                                    unresolved.append(file)
                        except FileNotFoundError:
                            # skip deleted files
                            continue
                    if unresolved:
                        if InputHelper.get_message('There are still unresolved conflicts. '
                                                   'Do you want to skip this patch',
                                                   default_yes=False):
                            inapplicable = True
                        else:
                            cls.old_repo.index.reset(paths=unresolved)
                            unresolved.insert(0, '--')
                            cls.old_repo.git.checkout(*unresolved, conflict='diff3')
                            continue
            if inapplicable:
                inapplicable_patches.append(patch_name)
                try:
                    cls.output_data = cls.old_repo.git.rebase(skip=True, stdout_as_string=True)
                except git.GitCommandError as e:
                    ret_code = e.status
                    cls.output_data = e.stdout
                    continue
                else:
                    break
            diff = cls.old_repo.index.diff(cls.old_repo.commit())
            if diff:
                modified_patches.append(patch_name)
            if next_index < last_index:
                if not InputHelper.get_message('Do you want to continue with another patch'):
                    raise KeyboardInterrupt
            try:
                if diff:
                    cls.output_data = cls.old_repo.git.rebase('--continue', stdout_as_string=True)
                else:
                    cls.output_data = cls.old_repo.git.rebase(skip=True, stdout_as_string=True)
            except git.GitCommandError as e:
                ret_code = e.status
                cls.output_data = e.stdout
            else:
                break
        original_commits = list(cls.old_repo.iter_commits(rev=cls.old_repo.branches.master))
        commits = list(cls.old_repo.iter_commits())
        untouched_patches = []
        deleted_patches = []
        for patch in cls.patches:
            patch_name = patch.get_patch_name()
            original_commit = [c for c in original_commits if cls.extract_patch_name(c.message) == patch_name]
            commit = [c for c in commits if cls.extract_patch_name(c.message) == patch_name]
            if original_commit and commit:
                if patch_name not in modified_patches and compare_commits(original_commit[0], commit[0]):
                    untouched_patches.append(patch_name)
                else:
                    base_name = os.path.join(cls.kwargs['rebased_sources_dir'], patch_name)
                    if commit[0].summary == cls.decorate_patch_name(patch_name):
                        diff = cls.old_repo.git.diff(commit[0].parents[0], commit[0], stdout_as_string=False)
                    else:
                        diff = cls.old_repo.git.format_patch(commit[0], '-1',
                                                             stdout=True, no_numbered=True,
                                                             no_attach=True, stdout_as_string=False)
                        diff = cls.strip_patch_name(diff, patch_name)
                    with open(base_name, 'wb') as f:
                        f.write(diff)
                        f.write(b'\n')
                    if patch_name not in modified_patches:
                        modified_patches.append(patch_name)
            elif patch_name not in inapplicable_patches:
                deleted_patches.append(patch_name)
        if deleted_patches:
            patch_dictionary['deleted'] = deleted_patches
        if modified_patches:
            patch_dictionary['modified'] = modified_patches
        if inapplicable_patches:
            patch_dictionary['inapplicable'] = inapplicable_patches
        if untouched_patches:
            patch_dictionary['untouched'] = untouched_patches
        return patch_dictionary

    @classmethod
    def apply_old_patches(cls, source_dir):
        """Function applies a patch to a old/new sources"""
        for patch in cls.patches:
            logger.info("Applying patch '%s' to '%s'",
                        patch.get_patch_name(),
                        os.path.basename(source_dir))
            try:
                cls.apply_patch(cls.old_repo, patch)
            except git.GitCommandError as e:
                raise RuntimeError('Failed to patch old sources') from e
        # update repository state
        cls.old_repo.git.config('rebasehelper.state', 'PATCHES', local=True)

    @classmethod
    def init_git(cls, directory):
        """Function initialize old and new Git repository"""
        try:
            repo = git.Repo(directory)
            try:
                state = repo.git.config('rebasehelper.state', get=True, local=True)
            except git.GitCommandError:
                del repo
                # repository not created by us, remove the metadata
                shutil.rmtree(os.path.join(directory, '.git'))
            else:
                return repo, state
        except git.InvalidGitRepositoryError:
            pass
        repo = git.Repo.init(directory)
        state = 'INIT'
        repo.git.config('rebasehelper.state', state, local=True)
        repo.git.config('user.name', GitHelper.get_user(), local=True)
        repo.git.config('user.email', GitHelper.get_email(), local=True)
        # prevent git commands from launching an interactive editor
        repo.git.config('core.editor', 'true', local=True)
        repo.git.add(all=True)
        repo.index.commit('Initial commit', skip_hooks=True)
        return repo, state

    @classmethod
    def patch(cls, old_dir, new_dir, rest_sources, patches, **kwargs):
        """
        The function can be used for patching one
        directory against another
        """
        cls.kwargs = kwargs
        cls.old_sources = old_dir
        cls.new_sources = new_dir
        cls.output_data = None
        cls.rest_sources = rest_sources
        cls.patches = patches
        cls.non_interactive = kwargs.get('non_interactive')
        cls.favor_on_conflict = kwargs.get('favor_on_conflict')
        cls.old_repo, old_repo_state = cls.init_git(old_dir)
        if old_repo_state == 'INIT':
            cls.apply_old_patches(old_dir)
        cls.new_repo, _ = cls.init_git(new_dir)
        return cls._git_rebase()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import io
import os

import pytest  # type: ignore

from rebasehelper.helpers.download_helper import DownloadHelper, DownloadError


class TestDownloadHelper:
    """ DownloadHelper tests """

    def test_keyboard_interrupt_situation(self, monkeypatch):
        """
        Test that the local file is deleted in case KeyboardInterrupt is raised during the download
        """
        KNOWN_URL = 'https://ftp.isc.org/isc/bind9/9.10.4-P1/srcid'
        LOCAL_FILE = os.path.basename(KNOWN_URL)

        def interrupter():
            raise KeyboardInterrupt

        # make sure that some function call inside tha actual download section raises the KeyboardInterrupt exception.
        monkeypatch.setattr('time.time', interrupter)

        with pytest.raises(KeyboardInterrupt):
            DownloadHelper.download_file(KNOWN_URL, LOCAL_FILE)

        assert not os.path.exists(LOCAL_FILE)

    @pytest.mark.parametrize('total, downloaded, output', [
        (100, 25, '\r 25%[=======>                      ]    25.00   eta 00:00:30 '),
        (100.0, 25.0, '\r 25%[=======>                      ]    25.00   eta 00:00:30 '),
        (-1, 1024 * 1024, '\r    [    <=>                       ]     1.00M   in 00:00:10 '),
    ], ids=[
        'integer',
        'float',
        'unknown_size',
    ])
    def test_progress(self, total, downloaded, output, monkeypatch):
        """
        Test that progress of a download is shown correctly. Test the case when size parameters are passed as integers.
        """
        buf = io.StringIO()
        monkeypatch.setattr('sys.stdout', buf)
        monkeypatch.setattr('time.time', lambda: 10.0)
        DownloadHelper.progress(total, downloaded, 0.0)
        assert buf.getvalue() == output

    @pytest.mark.parametrize('url, content', [
        ('http://integration:8000/existing_file.txt', 'content'),
        ('https://integration:4430/existing_file.txt', 'content'),
        ('ftp://integration:2100/existing_file.txt', 'content'),
        ('http://integration:8001/existing_file.txt', 'content'),
        ('https://integration:4431/existing_file.txt', 'content'),
        ('ftp://integration:2101/existing_file.txt', 'content'),
    ], ids=[
        'HTTP',
        'HTTPS',
        'FTP',
        'HTTP-unknown_size',
        'HTTPS-unknown_size',
        'FTP-unknown_size',
    ])
    @pytest.mark.integration
    def test_download_existing_file(self, url, content):
        """Test downloading existing file"""
        local_file = 'local_file'
        DownloadHelper.download_file(url, local_file)
        assert os.path.isfile(local_file)
        with open(local_file) as f:
            assert f.readline().strip() == content

    @pytest.mark.parametrize('url', [
        'http://integration:8000/non_existing_file.txt',
        'https://integration:4430/non_existing_file.txt',
        'ftp://integration:2100/non_existing_file.txt',
    ], ids=[
        'HTTP',
        'HTTPS',
        'FTP',
    ])
    @pytest.mark.integration
    def test_download_non_existing_file(self, url):
        """Test downloading NON existing file"""
        local_file = 'local_file'
        with pytest.raises(DownloadError):
            DownloadHelper.download_file(url, local_file)
        assert not os.path.isfile(local_file)
